"""
All of the feedback responses generated by TIFA.
"""
from pedal.types.new_types import ModuleType, FunctionType, ClassType
from pedal.utilities.operators import OPERATION_DESCRIPTION
from pedal.core.report import MAIN_REPORT
from pedal.core.feedback import FeedbackResponse


class TifaFeedback(FeedbackResponse):
    """ Base class for all TIFA feedback """
    muted = False
    category = FeedbackResponse.CATEGORIES.ALGORITHMIC
    kind = FeedbackResponse.KINDS.MISTAKE


class action_after_return(TifaFeedback):
    """ Statement after return """
    title = "Action after Return"
    message_template = ("You performed an action after already returning from "
                        "a function, on line {location.line:line}. You can "
                        "only return on a path once.")
    justification = ("TIFA visited a node not in the top scope when its "
                     "*return variable was definitely set in this scope.")

    def __init__(self, location, **kwargs):
        super().__init__(location=location, **kwargs)


class return_outside_function(TifaFeedback):
    """ Return statement outside of function """
    title = "Return outside Function"
    message_template = ("You attempted to return outside of a function on line "
                        "{location.line:line}. But you can only return from within "
                        "a function.")
    justification = "TIFA visited a return node at the top level."

    def __init__(self, location, **kwargs):
        super().__init__(location=location, **kwargs)


class multiple_return_types(TifaFeedback):
    """ Multiple returned types in single function """
    title = "Multiple Return Types"
    message_template = ("Your function returned {actual} on line {location.line:line}, "
                        "even though you defined it to return {expected}. "
                        "Your function should return values consistently.")
    justification = ("TIFA visited a function definition with multiple returns "
                     "that unequal types.")

    def __init__(self, location, expected, actual, **kwargs):
        super().__init__(location=location, expected=expected,
                         actual=actual, **kwargs)


class write_out_of_scope(TifaFeedback):
    """ Write out of Scope """
    title = "Write Out of Scope"
    message_template = ("You attempted to write the variable {name_message} "
                        "from a higher scope (outside the function) on line "
                        "{location.line:line}. You should only use variables inside "
                        "the function they were declared in.")
    justification = "TIFA stored to an existing variable not in this scope"

    def __init__(self, location, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        super().__init__(location=location, name=name,
                         name_message=report.format.name(name), **kwargs)


class unconnected_blocks(TifaFeedback):
    """ Unconnected Blocks """
    title = "Unconnected Blocks"
    message_template = ("It looks like you have unconnected blocks on line {location.line:line}. "
                        "Before you run your program, you must make sure that all "
                        "of your blocks are connected that there are no unfilled "
                        "holes.")
    justification = "TIFA found a name equal to ___"

    def __init__(self, location, **kwargs):
        super().__init__(location=location, **kwargs)


class iteration_problem(TifaFeedback):
    """ Iteration Problem """
    title = "Iteration Problem"
    message_template = ("The variable {name_message} was iterated on line "
                        "{location.line:line} but you used the same variable as the iteration "
                        "variable. You should choose a different variable name "
                        "for the iteration variable. Usually, the iteration variable "
                        "is the singular form of the iteration list (e.g., "
                        "`for a_dog in dogs:`).")
    justification = "TIFA visited a loop where the iteration list and target were the same."

    def __init__(self, location, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        super().__init__(location=location, name=name,
                         name_message=report.format.name(name), **kwargs)


class initialization_problem(TifaFeedback):
    """ Initialization Problem """
    title = "Initialization Problem"
    message_template = ("The variable {name_message} was used on line {location.line:line}, "
                        "but it was not given a value on a previous line. "
                        "You cannot use a variable until it has been given a value."
                        )
    justification = "TIFA read a variable that did not exist or was not previously set in this branch."

    def __init__(self, location, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        super().__init__(location=location, name=name,
                         name_message=report.format.name(name), **kwargs)


class possible_initialization_problem(TifaFeedback):
    """ Possible Initialization Problem """
    title = "Possible Initialization Problem"
    message_template = ("The variable {name_message} was used on line {location.line:line}, "
                        "but it was possibly not given a value on a previous "
                        "line. You cannot use a variable until it has been given "
                        "a value. Check to make sure that this variable was "
                        "declared in all of the branches of your decision."
                        )
    justification = "TIFA read a variable that was maybe set but not definitely set in this branch."

    def __init__(self, location, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        super().__init__(location=location, name=name,
                         name_message=report.format.name(name), **kwargs)


class unused_variable(TifaFeedback):
    """ Unused Variable """
    title = "Unused Variable"
    message_template = ("The {kind} {name_message} was {initialization} on line "
                        "{location.line:line}, but was never used after that.")
    justification = ("TIFA stored a variable but it was not read any other time "
                     "in the program.")

    def __init__(self, location, name, variable_type, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        if isinstance(variable_type, ModuleType):
            kind, initialization = 'module', 'imported'
        elif isinstance(variable_type, FunctionType):
            kind, initialization = 'function', 'given a definition'
        elif isinstance(variable_type, ClassType):
            kind, initialization = 'class', 'given a definition'
        # TODO: Handle classes... anything else?
        else:
            kind, initialization = 'variable', 'given a value'
        fields = {'location': location, 'name': name, 'type': variable_type,
                  'name_message': report.format.name(name),
                  'kind': kind, 'initialization': initialization}
        if 'fields' in kwargs:
            fields.update(kwargs.pop('fields'))
        super().__init__(location=location, fields=fields, **kwargs)


class overwritten_variable(TifaFeedback):
    """ Overwritten Variable """
    title = "Overwritten Variable"
    message_template = ("The variable {name_message} was given a value, but "
                        "{name_message} was changed on line {location.line:line} "
                        "before it was used. One of the times that you gave "
                        "{name_message} a value was incorrect."
                        )
    justification = ("TIFA attempted to store to a variable that was previously "
                     "stored but not read.")

    def __init__(self, location, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        super().__init__(location=location, name=name,
                         name_message=report.format.name(name), **kwargs)


class iterating_over_non_list(TifaFeedback):
    """ Iterating over non-list """
    title = "Iterating over Non-list"
    message_template = ("The {iter} is not a list, but you used it in the "
                        "iteration on line {location.line:line}. You should only "
                        "iterate over sequences like lists.")
    justification = ("TIFA visited a loop's iteration list whose type was"
                     "not indexable.")

    def __init__(self, location, iter_name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        if iter_name is None:
            iter_list = "expression"
        else:

            iter_list = "variable " + report.format.name(iter_name)
        fields = {'location': location, 'name': iter_name, 'iter': iter_list}
        if 'fields' in kwargs:
            fields.update(kwargs.pop('fields'))
        super().__init__(location=location, fields=fields, **kwargs)


class iterating_over_empty_list(TifaFeedback):
    """ Iterating over empty list """
    title = "Iterating over empty list"
    message_template = ("The {iter} was set as an empty list, "
                        "and then you attempted to use it in an iteration on line "
                        "{location.line:line}. You should only iterate over non-empty lists."
                        )
    justification = "TIFA visited a loop's iteration list that was empty."

    def __init__(self, location, iter_name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        if iter_name is None:
            iter_list = "expression"
        else:
            iter_list = "variable " + report.format.name(iter_name)
        fields = {'location': location, 'name': iter_name, 'iter': iter_list}
        super().__init__(location=location, fields=fields, **kwargs)


class incompatible_types(TifaFeedback):
    """ Incompatible types """
    title = "Incompatible types"
    message_template = ("You used {op_name} operation with {left_name} and {right_name} on line "
                        "{location.line:line}. But you can't do that with that operator. Make "
                        "sure both sides of the operator are the right type."
                        )
    justification = "TIFA visited an operation with operands of the wrong type."

    def __init__(self, location, operation, left, right, **kwargs):
        op_name = OPERATION_DESCRIPTION.get(operation.__class__,
                                            str(operation))
        left_name = left.singular_name
        right_name = right.singular_name
        fields = {'location': location,
                  'operation': operation, 'op_name': op_name,
                  'left': left, 'right': right,
                  'left_name': left_name, 'right_name': right_name}
        super().__init__(location=location, fields=fields, **kwargs)


class invalid_indexing(TifaFeedback):
    """ Invalid Index """
    title = "Invalid Index"
    message_template = ("You indexed {left_name} with {right_name} on line "
                        "{location.line:line}. But you can't index {left_name} with "
                        "{right_name}."
                        )
    justification = ("TIFA attempted to call an .index() operation on a type"
                     " with a type that wasn't acceptable.")
    muted = True

    def __init__(self, location, left, right, **kwargs):
        left_name = left.singular_name
        right_name = right.singular_name
        fields = {'location': location,
                  'left': left, 'right': right,
                  'left_name': left_name, 'right_name': right_name}
        super().__init__(location=location, fields=fields, **kwargs)


class parameter_type_mismatch(TifaFeedback):
    """ Parameter type mismatch """
    title = "Parameter Type Mismatch"
    message_template = ("You defined the parameter {parameter_name_message} "
                        "as {parameter_type_name}. However, the argument passed to that parameter on line {location.line:line} "
                        "was {argument_type_name}. The formal parameter type must match the argument's type."
                        )
    justification = "TIFA visited a function definition where a parameter type and argument type were not equal."

    def __init__(self, location, parameter_name, parameter, argument, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        parameter_type_name = parameter.singular_name
        argument_type_name = argument.singular_name
        if parameter_type_name == argument_type_name:
            argument_type_name = "a different kind of " + argument_type_name
        fields = {'location': location,
                  'parameter_name': parameter_name,
                  'parameter_name_message': report.format.name(parameter_name),
                  'parameter_type': parameter,
                  'argument_type': argument,
                  'parameter_type_name': parameter_type_name,
                  'argument_type_name': argument_type_name}
        super().__init__(location=location, fields=fields, **kwargs)


class field_type_mismatch(TifaFeedback):
    """ Field type mismatch """
    title = "Field Type Mismatch"
    message_template = ("You defined the field {field_name_message} "
                        "as {field_type_name}. However, the argument passed to that field's parameter in the"
                        " constructor function on line {location.line:line} was {argument_type_name}. The formal field"
                        " type must match the argument's type."
                        )
    justification = "TIFA visited a constructor function call where a parameter type and argument type were not subtypes."

    def __init__(self, location, field_name, field, argument, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        field_type_name = field.singular_name
        argument_type_name = argument.singular_name
        if field_type_name == argument_type_name:
            argument_type_name = "a different kind of " + argument_type_name
        fields = {'location': location,
                  'field_name': field_name,
                  'field_name_message': report.format.name(field_name),
                  'field_type': field,
                  'field_type_name': field_type_name,
                  'argument_type': argument,
                  'argument_type_name': argument_type_name}
        super().__init__(location=location, fields=fields, **kwargs)

class read_out_of_scope(TifaFeedback):
    """ Read out of scope """
    title = "Read out of Scope"
    message_template = ("You attempted to read the variable {name_message} "
                        "from a different scope on line {location.line:line}. You "
                        "should only use variables inside the function they "
                        "were declared in."
                        )
    justification = "TIFA read a variable that did not exist in this scope but existed in another."

    def __init__(self, location, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        super().__init__(location=location, name=name,
                         name_message=report.format.name(name), **kwargs)


# TODO: Complete these
class type_changes(TifaFeedback):
    """ Type changes """
    title = "Type Changes"
    message_template = ("The variable {name_message} changed type from {old} to "
                        "{new} on line {location.line:line}.")
    justification = ""
    muted = True

    def __init__(self, location, name, old, new, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = {'location': location, 'name': name,
                  'name_message': report.format.name(name),
                  'old': old, 'new': new}
        super().__init__(location=location, fields=fields, **kwargs)


class attribute_type_change(TifaFeedback):
    """ Type change on attribute """
    title = "Attribute Type Change"
    message_template = ("On line {location.line:line}, you attempted to assign a {new} to "
                        "the attribute {attr_message}, which was supposed to be {old}.")
    justification = ""
    muted = False

    def __init__(self, location, attr, old, new, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = {'location': location, 'attr': attr,
                  'attr_message': report.format.name(attr),
                  'old': old, 'new': new}
        super().__init__(location=location, fields=fields, **kwargs)

class type_change_append(TifaFeedback):
    """ Type Change in an append Method """
    title = "Type Change in Append"
    message_template = ("You attempted to append a {new} to a list that was "
                        "supposed to have {old} on line {location.line:line}.")
    justification = ""
    muted = True

    def __init__(self, location, old, new, **kwargs):
        fields = {'location': location, 'old': old, 'new': new}
        super().__init__(location=location, fields=fields, **kwargs)


class unnecessary_second_branch(TifaFeedback):
    """ Unnecessary second branch """
    title = "Unnecessary Second Branch"
    message_template = ("You have an `if` statement where one of the two branches"
                       " only has `pass` in its body, on line {location.line:line}."
                        " You shouldn't need an empty body.")
    justification = "There is an else or if statement who's body is just pass."

    def __init__(self, location, **kwargs):
        super().__init__(location=location, **kwargs)


class else_on_loop_body(TifaFeedback):
    """ Else on Loop body """
    title = "Else on Loop Body"
    message_template = "TODO"
    justification = ""

    def __init__(self, location, **kwargs):
        super().__init__(location=location, **kwargs)


class recursive_call(TifaFeedback):
    """ recursive call """
    title = "Recursive Call"
    message_template = "TODO"
    justification = ""
    muted = True

    def __init__(self, location, name, **kwargs):
        super().__init__(location=location, name=name, **kwargs)


class not_a_function(TifaFeedback):
    """ Not a function """
    title = "Not a Function"
    message_template = ("You attempted to call {name} as if it"
                        " was a function on line {location.line:line}. However,"
                        " that expression was actually {singular_name}.")
    justification = ""
    # TODO: Unmute?
    #muted = True

    def __init__(self, location, name, called_type, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        singular_name = called_type.singular_name
        fields = {'location': location, 'name': name,
                  'called_type': called_type,
                  'singular_name': singular_name}
        super().__init__(fields=fields, **kwargs)


class incorrect_arity(TifaFeedback):
    """ Incorrect arity """
    title = "Incorrect Arity"
    message_template = ("The {function_type} {function_name_message} was given the "
                        "wrong number of arguments. You should have had {expected_count} "
                        "arguments, but instead you had {actual_count} arguments.")
    justification = ""

    def __init__(self, location, function_name, expected_count, actual_count, is_constructor=False, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        function_type = 'constructor function' if is_constructor else 'function'
        super().__init__(location=location, function_name=function_name,
                         expected_count=expected_count, actual_count=actual_count,
                         function_name_message=report.format.name(function_name),
                         function_type=function_type,
                         **kwargs)


class module_not_found(TifaFeedback):
    """ Module not found """
    title = "Module Not Found"
    message_template = "TODO"
    justification = ""
    muted = True

    def __init__(self, location, name, is_dynamic=False, error=None, **kwargs):
        fields = {"location": location, "name": name,
                  "is_dynamic": is_dynamic, "error": error}
        super().__init__(location=location, fields=fields, **kwargs)


class append_to_non_list(TifaFeedback):
    """ Append to non-list """
    title = "Append to non-list"
    message_template = "TODO"
    justification = ""
    muted = True

    def __init__(self, location, name, actual_type, **kwargs):
        fields = {'location': location, "name": name,
                  "actual_type": actual_type}
        super().__init__(location=location, fields=fields, **kwargs)


class nested_function_definition(TifaFeedback):
    """ Function defined not at top-level """
    message_template = ("The function {name_message} was defined inside of another"
                        "block on line {location.line:line}. For instance, you may "
                        "have placed it inside another function definition, or "
                        "inside of a loop. Do not nest your function "
                        "definition!")
    title = "Don't Nest Functions"
    justification = "Found a FunctionDef that was not at the top-level."
    muted = True
    unscored = True

    def __init__(self, location, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        super().__init__(location=location, name=name,
                         name_message=report.format.name(name), **kwargs)


class unused_returned_value(TifaFeedback):
    """ Expr node had a non-None value """
    title = "Did Not Use Function's Return Value"
    message_template = ("It looks like you called the {call_type} {name_message} on "
                        "{location.line:line}, but failed to store the result in "
                        "a variable or use it in an expression. You should "
                        "remember to use the result!")
    justification = "Expression node calculated a non-None value."
    muted = True
    unscored = True

    def __init__(self, location, name, call_type, result_type, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = {'location': location, 'name': name, 'call_type': call_type,
                  'result_type': result_type,
                  'name_message': report.format.name(name)}
        super().__init__(fields=fields, location=location, **kwargs)


'''
TODO: Finish these checks
"Empty Body": [], # Any use of pass on its own
"Malformed Conditional": [], # An if/else with empty else or if
"Unnecessary Pass": [], # Any use of pass
"Append to non-list": [], # Attempted to use the append method on a non-list
"Used iteration list": [], #
"Unused iteration variable": [], #
"Type changes": [], #
"Unknown functions": [], #
"Not a function": [], # Attempt to call non-function as function
"Recursive Call": [],
"Incorrect Arity": [],
"Aliased built-in": [], #
"Method not in Type": [], # A method was used that didn't exist for that type
"Submodule not found": [],
"Module not found": [],
"Else on loop body": [], # Used an Else on a For or While
'''


# TODO: Add in all other names from this module
FEEDBACK_BY_NAME = {'type_change_append': type_change_append,
                    'incorrect_arity': incorrect_arity,
                    'parameter_type_mismatch': parameter_type_mismatch,
                    'field_type_mismatch': field_type_mismatch}


def lookup_feedback(name):
    return FEEDBACK_BY_NAME.get(name)



"""
Utilities related to more sophisticated sorting methods.
"""


def dfs(name: str, visited, result, orderings):
    """
    Recursive Depth-first-search helper function that will explore one level
    of depth from a given node.

    Args:
        name (str): The starting point to DFS.
        visited (set[str]): The previously visited nodes.
        result (list[str]): The currently produced sorted list of nodes.
        orderings (dict[str, list[str]]): Dictionary that maps each node to
            the list of nodes that they must precede.
    """
    visited.add(name)
    if name in orderings:
        for neighbor in orderings[name]:
            if neighbor not in visited:
                dfs(neighbor, visited, result, orderings)
    result.insert(0, name)


def topological_sort(names, orderings):
    """
    Given a collection of strings, perform a Depth-first-search to topologically
    sort them. Respects any ``orderings`` that are given.

    Args:
        names (list[str]): The complete list of all names that are to be
            sorted.
        orderings (dict[str, list[str]]): Dictionary that maps each node to
            the list of nodes that they must precede.

    Returns:
        list[str]: The ordered list of strings, respecting the topology.
    """
    visited = set()
    result = []
    for name in names[::-1]:
        if name not in visited:
            dfs(name, visited, result, orderings)
    return result



"""
Main TIFA visitor-based algorithm here.

TODO: JoinedStr
"""

import ast
from pedal.core.report import MAIN_REPORT
from pedal.core.location import Location
from pedal.tifa.contexts import NewScope
from pedal.types.normalize import (get_pedal_type_from_json)
from pedal.types.builtin import (get_builtin_module)
from pedal.types.new_types import (is_subtype, Type, AnyType, ImpossibleType, NoneType,
                                   ListType, TupleType, ModuleType,
                                   LiteralStr, LiteralInt, LiteralFloat, LiteralBool)
from pedal.tifa.constants import TOOL_NAME
from pedal.tifa.identifier import Identifier
from pedal.tifa.state import State
from pedal.tifa.feedbacks import (action_after_return, return_outside_function,
                                  write_out_of_scope, unconnected_blocks,
                                  iteration_problem, not_a_function,
                                  initialization_problem, unused_variable,
                                  overwritten_variable, iterating_over_non_list,
                                  iterating_over_empty_list, incompatible_types,
                                  parameter_type_mismatch, read_out_of_scope,
                                  type_changes, unnecessary_second_branch,
                                  recursive_call, multiple_return_types,
                                  possible_initialization_problem,
                                  incorrect_arity, else_on_loop_body,
                                  module_not_found, nested_function_definition,
                                  unused_returned_value, invalid_indexing, append_to_non_list,
                                  lookup_feedback)

__all__ = ['TifaCore', 'TifaAnalysis']


class TifaAnalysis:
    """
    Container class for the results of running a Tifa Analysis.

    Attributes:
        success (bool): Whether or not the analysis was able to finish.
        variables (dict): 2D Dictionary mapping (Scope Names x Variable Names)
            to their calculated types. This captures ALL the variables in
            the code (in all scopes), as opposed to just the top level
            variables.
        top_level_variables (dict): Maps variable names to their calculated
            types.
        node_types (dict): Maps AST nodes to their TIFA-calculated type.
    """
    def __init__(self):
        self.success = True
        self.error = None
        self.variables = {}
        self.top_level_variables = {}
        self.issues = {}
        self.node_types = {}

    def fail(self, error):
        """ Called when this analysis failed. """
        self.success = False
        self.error = error


class TifaCore:
    """
    TIFA Core, with methods for manipulating the TifaAnalysis data.

    Args:
        report (Report): The report object to store data and feedback in. If
                         left None, defaults to the global MAIN_REPORT.
    """
    path_id: int
    scope_id: int
    ast_id: int

    path_names: list
    scope_names: list

    node_chain: list
    scope_chain: list
    path_chain: list
    definition_chain: list

    # dict[paths, dict[names, State]]
    name_map: dict
    class_scopes: dict
    path_parents: dict

    final_node: ast.AST or None

    line_offset: int

    def __init__(self, report=MAIN_REPORT):
        self.report = report
        self.analysis = None

    def get_setting(self, setting_name):
        return self.report[TOOL_NAME]['settings'][setting_name]

    def locate(self, node: ast.AST = None):
        """
        Return a dictionary representing the current location within the
        AST.

        Returns:
            Location: The line and column of the current or given node.
        """
        if node is None:
            if self.node_chain:
                node = self.node_chain[-1]
            else:
                node = self.final_node
        return Location(node.lineno+self.line_offset, col=node.col_offset)

    def _issue(self, *feedback):
        if len(feedback) == 1:
            feedback = feedback[0]
        else:
            feedback = lookup_feedback(feedback[0])(*feedback[1:])
        if feedback.label not in self.analysis.issues:
            self.analysis.issues[feedback.label] = []
        self.analysis.issues[feedback.label].append(feedback)

    def _collect_top_level_variables(self):
        """
        Walk through the variables and add any at the top level to the
        top_level_variables field of the report.
        """
        top_level_variables = self.analysis.top_level_variables
        main_path_vars = self.name_map[self.path_chain[0]]
        for full_name in main_path_vars:
            split_name = full_name.split("/")
            if len(split_name) == 2 and split_name[0] == str(self.scope_chain[0]):
                name = split_name[1]
                top_level_variables[name] = main_path_vars[full_name]

    def reset(self):
        """
        Reinitialize fields for maintaining the system
        """
        # Unique Global IDs
        self.path_id = 0
        self.scope_id = 0
        self.ast_id = 0

        # Human readable names
        self.path_names = ['*Module']
        self.scope_names = ['*Module']
        self.node_chain = []

        # Complete record of all Names
        self.scope_chain = [self.scope_id]
        self.path_chain = [self.path_id]
        self.name_map = {self.path_id: {}}
        self.definition_chain = []
        self.path_parents = {}
        self.final_node = None
        self.class_scopes = {}
        self.module_scopes = {}

    def find_variable_scope(self, name):
        """
        Walk through this scope and all enclosing scopes, finding the relevant
        identifier given by `name`.

        Args:
            name (str): The name of the variable
        Returns:
            Identifier: An Identifier for the variable, which could potentially
                        not exist.
        """
        for scope_level, scope in enumerate(self.scope_chain):
            for path_id in self.path_chain:
                path = self.name_map[path_id]
                full_name = "/".join(map(str, self.scope_chain[scope_level:])) + "/" + name
                if full_name in path:
                    is_root_scope = (scope_level == 0)
                    return Identifier(True, is_root_scope,
                                      full_name, path[full_name])

        return Identifier(False)

    def find_variable_out_of_scope(self, name):
        """
        Walk through every scope and determine if this variable can be found
        elsewhere (which would be an issue).

        Args:
            name (str): The name of the variable
        Returns:
            Identifier: An Identifier for the variable, which could potentially
                        not exist.
        """
        for path in self.name_map.values():
            for full_name in path:
                unscoped_name = full_name.split("/")[-1]
                if name == unscoped_name:
                    return Identifier(True, False, unscoped_name, path[full_name])
        return Identifier(False)

    def find_path_parent(self, path_id, name):
        """

        Args:
            path_id:
            name:

        Returns:

        """
        if name in self.name_map[path_id]:
            return Identifier(True, state=self.name_map[path_id][name])
        else:
            path_parent = self.path_parents.get(path_id)
            if path_parent is None:
                return Identifier(False)
            else:
                return self.find_path_parent(path_parent, name)

    def _finish_scope(self):
        """
        Walk through all the variables present in this scope and ensure that
        they have been read and not overwritten.
        """
        path_id = self.path_chain[0]
        for name in self.name_map[path_id]:
            if self.in_scope(name, self.scope_chain):
                state = self.name_map[path_id][name]
                if state.over == 'yes':
                    position = state.over_position
                    self._issue(overwritten_variable(position, state.name))
                if state.read == 'no' and state.name != '_':
                    self._issue(unused_variable(state.position, state.name, state.type, report=self.report))

    def _scope_chain_str(self, name=None):
        """
        Convert the current scope chain to a string representation (divided 
        by "/").

        Returns:
            str: String representation of the scope chain.
        """
        if name:
            return "/".join(map(str, self.scope_chain)) + "/" + name
        else:
            return "/".join(map(str, self.scope_chain))

    def identify_caller(self, node):
        """
        Figures out the variable that was used to kick off this call,
        which is almost always the relevant Name to track as being updated.
        If the origin wasn't a Name, nothing will need to be updated so None
        is returned instead.

        TODO: Is this sufficient?

        Args:
            node (AST): An AST node
        Returns:
            str or None: The name of the variable or None if no origin could
                         be found.
        """
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Call):
            return self.identify_caller(node.func)
        elif isinstance(node, (ast.Attribute, ast.Subscript)):
            return self.identify_caller(node.value)
        return None

    def iterate_variable(self, name, position=None):
        """
        Update the variable by iterating through it - this doesn't do anything
        fancy yet.
        """
        return self.load_variable(name, position)

    def store_read_variable(self, name, new_type, position=None):
        """
        Record that the variable was iterated (or otherwise stored in a way that is safe
        to not read later on) upon. This counts as a Read. Another use case is class variables.

        Args:
            name (str): The name of the variable.
            new_type (Type): The Tifa Type of the variable.
            position: The location where this iteration occurred.

        Returns:
            :py:class:`pedal.tifa.state.State`: The wrapped State object of
                the stored variable.
        """
        state = self.store_variable(name, new_type, position)
        state.read = 'yes'
        return state

    def return_variable(self, return_type):
        """

        Args:
            return_type:

        Returns:

        """
        return self.store_variable("*return", return_type)

    def append_variable(self, name, append_type, position=None):
        """

        Args:
            name:
            append_type:
            position:

        Returns:

        """
        return self.store_variable(name, append_type, position)

    def create_variable(self, name, create_type, position=None):
        """ Stores a new variable, but unconditionally - even if it exists
        in another scope, it will be considered a new variable."""
        return self.store_variable(name, create_type, position, force_create=True)

    def store_variable(self, name, store_type, position=None, force_create=False):
        """
        Update the variable with the given name to now have the new type.

        Args:
            name (str): The unqualified name of the variable. The variable will
                        be assumed to be in the current scope.
            store_type (Type): The new type of this variable.
            position: The location that this store occurred at.
            force_create: Used to force the creation of a new variable, even if the variable already exists.
        Returns:
            State: The new state of the variable.
        """
        if position is None:
            position = self.locate()
        full_name = self._scope_chain_str(name)
        current_path = self.path_chain[0]
        variable = self.find_variable_scope(name)
        if not variable.exists or force_create:
            # Create a new instance of the variable on the current path
            new_state = State(name, [], store_type, 'store', position,
                              read='no', set='yes', over='no')
            self.name_map[current_path][full_name] = new_state
        else:
            new_state = self.trace_state(variable.state, "store", position)
            if not variable.in_scope and not self._in_module():
                self._issue(write_out_of_scope(self.locate(), name, report=self.report))
            # Type change?
            if not is_subtype(store_type, variable.state.type):
                self._issue(type_changes(position, name, variable.state.type, store_type))
            new_state.type = store_type
            # Overwritten?
            if variable.state.set == 'yes' and variable.state.read == 'no':
                new_state.over_position = position
                new_state.over = 'yes'
            else:
                new_state.set = 'yes'
                new_state.read = 'no'
            self.name_map[current_path][full_name] = new_state
        # If this is a class scope...
        current_scope = self.scope_chain[0]
        if current_scope in self.class_scopes:
            self.class_scopes[current_scope].add_attr(name, new_state.type)
        return new_state

    def load_variable(self, name, position=None):
        """
        Retrieve the variable with the given name.

        Args:
            position:
            name (str): The unqualified name of the variable. If the variable is
                        not found in the current scope or an enclosing scope, all
                        other scopes will be searched to see if it was read out
                        of scope.
        Returns:
            State: The current state of the variable.
        """
        full_name = self._scope_chain_str(name)
        current_path = self.path_chain[0]
        variable = self.find_variable_scope(name)
        if position is None:
            position = self.locate()
        if not variable.exists:
            out_of_scope_var = self.find_variable_out_of_scope(name)
            # Create a new instance of the variable on the current path
            if out_of_scope_var.exists:
                self._issue(read_out_of_scope(self.locate(), name))
            else:
                self._issue(initialization_problem(self.locate(), name))
            new_state = State(name, [], AnyType(), 'load', position,
                              read='yes', set='no', over='no')
            self.name_map[current_path][full_name] = new_state
        else:
            new_state = self.trace_state(variable.state, "load", position)
            if variable.state.set == 'no':
                self._issue(initialization_problem(self.locate(), name))
            if variable.state.set == 'maybe':
                if name != '*return':
                    self._issue(possible_initialization_problem(self.locate(), name))
            new_state.read = 'yes'
            if not variable.in_scope:
                self.name_map[current_path][variable.scoped_name] = new_state
            else:
                self.name_map[current_path][full_name] = new_state
        return new_state

    def load_module(self, chain):
        """
        Finds the module in the set of available modules.

        Args:
            chain (str): A chain of module imports (e.g., "matplotlib.pyplot")
        Returns:
            ModuleType: The specific module with its members, or an empty
                        module type.
        """
        module_names = chain.split('.')
        for potential_module in [self.report[TOOL_NAME]['types']['modules'].get(module_names[0]),
                                 get_builtin_module(module_names[0])]:
            if potential_module is not None:
                base_module = potential_module
                for module in module_names[1:]:
                    if (isinstance(base_module, ModuleType) and
                            module in base_module.submodules):
                        base_module = base_module.submodules[module]
                    else:
                        # TODO: What if the module is partially overriding a builtin?
                        self._issue(module_not_found(self.locate(), chain, False, None, report=self.report))
                return base_module

        # Non-student file, maybe it has _tifa_definitions?
        error = None
        try:
            actual_module = __import__(chain, globals(), {},
                                       ['_tifa_definitions'])
            definitions = actual_module._tifa_definitions()
            return get_pedal_type_from_json(definitions)
        except Exception as e:
            error = e
        filename = chain.replace('.', '/') + ".py"
        if self.report.submission and filename in self.report.submission.files:
            # TODO: Try running TIFA over the code
            code = self.report.submission.files[filename]
            try:
                # TODO: Doesn't import toplevels as variables.
                # TODO: Allow variables to be unused inside the file
                self.report[TOOL_NAME]['types']['modules'][chain] = self._visit_module(chain, filename, code)
                return self.report[TOOL_NAME]['types']['modules'][chain]
            except Exception as e:
                error = e
        self._issue(module_not_found(self.locate(), chain, True, error, report=self.report))
        return ModuleType(module_names[-1], {}, {})

    def _visit_module(self, module_name, filename, code):
        ast_tree = ast.parse(code, filename)
        # TODO: Properly handle submodules
        new_module = ModuleType(module_name, {}, {})
        #self.store_variable(class_name, new_class_type)
        definitions_scope = self.scope_chain[:]
        class_scope = NewScope(self, definitions_scope, class_type=new_module, is_module=True)
        with class_scope:
            self.visit(ast_tree)
        return new_module

    def _in_module(self):
        return any(scope_id in self.module_scopes for scope_id in self.scope_chain)

    def combine_states(self, left, right):
        """

        Args:
            left:
            right:

        Returns:

        """
        state = State(left.name, [left], left.type, 'branch', self.locate(),
                      read=left.read, set=left.set, over=left.over,
                      over_position=left.over_position)
        if right is None:
            state.read = 'no' if left.read == 'no' else 'maybe'
            state.set = 'no' if left.set == 'no' else 'maybe'
            state.over = 'no' if left.over == 'no' else 'maybe'
        else:
            if not is_subtype(left.type, right.type):
                self._issue(type_changes(self.locate(), left.name, left.type, right.type))
            state.read = self.match_rso(left.read, right.read)
            state.set = self.match_rso(left.set, right.set)
            state.over = self.match_rso(left.over, right.over)
            if left.over == 'no':
                state.over_position = right.over_position
            state.trace.append(right)
        return state

    def merge_paths(self, parent_path_id, left_path_id, right_path_id):
        """
        Combines any variables on the left and right path into the parent
        name space.

        Args:
            parent_path_id (int): The parent path of the left and right branches
            left_path_id (int): One of the two paths
            right_path_id (int): The other of the two paths.
        """
        # Combine two paths into one
        # Check for any names that are on the IF path
        for left_name in self.name_map[left_path_id]:
            left_state = self.name_map[left_path_id][left_name]
            right_identifier = self.find_path_parent(right_path_id, left_name)
            if right_identifier.exists:
                # Was on both IF and ELSE path
                right_state = right_identifier.state
            else:
                # Was only on IF path, potentially on the parent path
                right_state = self.search_parents(parent_path_id, left_name)
                # right_state = self.name_map[parent_path_id].get(left_name)
            combined = self.combine_states(left_state, right_state)
            self.name_map[parent_path_id][left_name] = combined
        # Check for names that are on the ELSE path but not the IF path
        for right_name in self.name_map[right_path_id]:
            if right_name not in self.name_map[left_path_id]:
                right_state = self.name_map[right_path_id][right_name]
                # Potentially on the parent path
                #parent_state = self.name_map[parent_path_id].get(right_name)
                parent_state = self.search_parents(parent_path_id, right_name)
                combined = self.combine_states(right_state, parent_state)
                self.name_map[parent_path_id][right_name] = combined

    def search_parents(self, parent_id, seeking_name):
        possible = self.name_map[parent_id]
        if seeking_name in possible:
            return possible[seeking_name]
        elif parent_id in self.path_parents:
            parent_id = self.path_parents[parent_id]
            return self.search_parents(parent_id, seeking_name)
        else:
            return None

    @staticmethod
    def trace_state(state, method, position):
        """
        Makes a copy of the given state with the given method type.

        Args:
            state (State): The state to copy (as in, we trace a copy of it!)
            method (str): The operation being applied to the state.
            position: The location this copy occurred at.
        Returns:
            State: The new State
        """
        return state.copy(method, position)

    @staticmethod
    def in_scope(full_name, scope_chain):
        """
        Determine if the fully qualified variable name is in the given scope
        chain.

        Args:
            full_name (str): A fully qualified variable name
            scope_chain (list): A representation of a scope chain.
        Returns:
            bool: Whether the variable lives in this scope
        """
        # Get this entity's full scope chain
        name_scopes = full_name.split("/")[:-1]
        # against the reverse scope chain
        checking_scopes = [str(s) for s in scope_chain]
        return name_scopes == checking_scopes

    @staticmethod
    def match_rso(left, right):
        """

        Args:
            left:
            right:

        Returns:

        """
        if left == right:
            return left
        else:
            return "maybe"

    def get_literal(self, node):
        """

        Args:
            node:

        Returns:

        """
        if isinstance(node, ast.Num):
            return LiteralInt(node.n) if isinstance(node.n, int) else LiteralFloat(node.n)
        elif isinstance(node, ast.Str):
            return LiteralStr(node.s)
        elif isinstance(node, ast.Tuple):
            values = []
            for elt in node.elts:
                subvalue = self.get_literal(elt)
                if subvalue is not None:
                    values.append(subvalue)
                else:
                    return None
            return TupleType(values)
        elif isinstance(node, ast.Name):
            if node.id == "None":
                return NoneType()
            elif node.id == "False":
                return LiteralBool(False)
            elif node.id == "True":
                return LiteralBool(True)
        return None

    def check_common_bad_lookups(self, left_type, attr, node):
        """ Handle common kinds of bad lookups, like appending to a non-list. """
        if attr == "append":
            if not isinstance(left_type, ListType):
                self._issue(append_to_non_list(self.locate(node),
                                               self.identify_caller(node),
                                               left_type,
                                               report=self.report))



"""
Check utilities.sorting
"""
import unittest
from pedal.utilities.sorting import topological_sort


class TestTopologicalSort(unittest.TestCase):
    def test_topological_sort(self):
        basic = ['Apple', 'Banana', 'Corn', 'Date', 'Eggplant']
        self.assertEqual(topological_sort(basic, {}), basic)

        backwards = basic[::-1]
        self.assertEqual(topological_sort(backwards, {}), backwards)

        force = {'Apple': ['Corn', 'Date'],
                 'Date': ['Corn', 'Eggplant'],
                 'Banana': ['Eggplant']}
        self.assertEqual(topological_sort(basic, force),
                         ['Apple', 'Banana', 'Date', 'Corn', 'Eggplant'])

        self.assertEqual(topological_sort(backwards, force),
                         ['Banana', 'Apple', 'Date', 'Eggplant', 'Corn'])

        force = {'Eggplant': ['Date'],
                 'Date': ['Corn'],
                 'Corn': ['Banana'],
                 'Banana': ['Apple']}
        self.assertEqual(topological_sort(basic, force),
                         backwards)

        self.assertEqual(topological_sort(backwards, force),
                         backwards)

    def test_project_names(self):
        names = ['records', 'render_introduction',
                 'create_world', 'render',
                 'get_options', 'choose', 'update',
                 'render_ending', 'win_and_lose_paths',
                 'conclusion']
        orderings = {'choose': ['win_and_lose_paths'],
                     'create_world': ['render',
                                      'get_options',
                                      'render_ending',
                                      'win_and_lose_paths'],
                     'get_options': ['choose', 'update', 'win_and_lose_paths'],
                     'render': ['win_and_lose_paths'],
                     'render_ending': ['win_and_lose_paths'],
                     'render_introduction': ['win_and_lose_paths'],
                     'update': ['win_and_lose_paths'],
                     'win_and_lose_paths': ['conclusion']}
        expected = ['records',
                    'render_introduction',
                    'create_world',
                    'render',
                    'get_options', 'choose',
                    'update',
                    'render_ending',
                    'win_and_lose_paths',
                    'conclusion']
        self.assertEqual(topological_sort(names, orderings), expected)


if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Mock module for BBC Microbit API.

Image formats:
    list[list[int]] (or any kind of 2d sequence)
    list[list[str]]

    A single string with lines separated by colons.
    A single string with lines separated by newlines.
    A single string without separators.

    Brightness values can be integers or characters
    You can use "?" as a "don't care" symbol for string representations.
    Equivalent values:
        `0`, ` `, `_`, `⬛`
        `9`, `*`, `■`, `█`, `🟥`, `⬜`, `X`, `x`

    You can also specify via a dictionary:
        Points (x, y) => Brightness value


microbit.display.scroll - in addition to capturing the screen history, also capture the message history
microbit.display.show - and in general here too, capture any strings being show so they can be queries.

assert_microbit_displayed(image)
assert_microbit_displayed(image1, image2, ...)
assert_microbit_displaying(image)

assert_microbit_printed(text)
assert_microbit_printed_contains(text)

Lingering questions:
* How do we want to simulate inputs (e.g., pins, buttons, etc.)? Some kind of Jest-like API?

```python
# Pressing the button prints its letter
assert_microbit_printed(None)
user.press_button_a()
assert_microbit_printed('a')
user.press_button_b()
assert_microbit_printed('b')

# Simulating Flappy Bit?
assert_microbit_displayed("<bits indicating the bird is initially center>")
user.tilt_up()
```

# Infinite loops

A major issue is infinite loops.
* In some places, the loops are simply triggered by keyword arguments. For example, the `microbit.display.show` function
    can take a `loop` boolean to "loop forever". They also have a `wait` keyword to block until the animation is finished.
    Otherwise the animation happens in the background (asynchronously, I assume?).
* A major pattern is to have students do all their work within a `while True:` loop, or other kind of "infinite" loop.

The async story in Skulpt is questionable. I honestly have to go research what the current state is. I mean, heck,
the async story in JavaScript is a little questionable at times. But I think there is definitely some potential here,
even if we had to do gross things like use `exec` and stuff. So executing is not off the table, I'm pretty sure.

The harder question is how we meaningfully simulate it from the instructor code. I think a lot of this will come down
to what the instructor actually wants to do.

### While->Function

A simple strategy for handling the `while True` loop pattern is for the autograder to transform all `while True` loops
into function definitions with `if` statements inside. An example:

```python
# answer.py
from microbit.display import show
counter = 0
while True:
    show(counter)
    counter += 1

# Becomes answer.py (transformed)
from microbit.display import show
counter = 0
def _microbit_main_loop():
    global counter
    if True:
        show(counter)
        counter += 1

# Then the instructor code can be free to "use" the student code however they want
# instructor.py
from pedal import *

run()
assert_microbit_displayed(None)
call('_microbit_main_loop')     # Perhaps aliased to step_student() ?
assert_microbit_displaying(1, message="The first iteration of your loop should display the number `1`!")
call('_microbit_main_loop')
assert_microbit_displaying(2)
```

This `while->function` approach has several nightmares attached:
* How do we correctly identify between the main loop and other loops? What about inner infinite loops?
    * Heuristics could go pretty far; most of the time, won't code fit a known pattern?
* What about global variables that are modified?
    * We can track global variables and inject the `global` keyword into the function
* What about errors and such in the original code?
    * Pedal already handles this with the `source` module! It can at least have an "offset" in the code. But I'm
        thinking this may also require having "skipped" lines and/or "remapped" lines. The former would let you mark
        lines as not counting towards the total, and the latter would explicitly change existing lines.
    * Unfortunately, we don't *actually* have a way to "modify" code in Pedal. But that's easily changed, I've always
        considered writing such a module. And we can dig around for a suitable existing one if we want.

Let's do a longer example where we grade the circle light game but with button presses involved. In this version,
the A button makes it run around the edge (otherwise it will just sit there). The B button ends the game.

```python
# Start off with light in topleft
run()
assert_microbit_displaying("9" + "0"*24)
# Idling for 3 ticks should do nothing
for _ in range(3):
    step()
    assert_microbit_displaying("9" + "0"*24)
# Pressing A should make it do a circle
press_a()
step(times=16+1)   # Has to run around the perimeter, or 4*4 steps, plus one in case they need the help
assert_microbit_displayed(*EDGE_STRINGS)
# Make sure it returned to the starting point
assert_microbit_displaying("9" + "0"*24)
# Pressing B should end the loop
press_b()
step()
assert_microbit_loop_ended()
```

The `step` function finds the main game loop, and executes it. That function will return whether the loop body
was actually executed. That information is also stored in such a way that the `assert_microbit_loop_ended` function
can query it.

### While->For

A somewhat related concept is the idea of transforming the `while` loop into a `for` loop (or perhaps even an
`async for` loop?). That `for` loop can have a generator with our own tooling equipped.

```python
# answer.py
from microbit.display import show
counter = 0
while True:
    show(counter)
    counter += 1

# Becomes
from microbit.display import show
counter = 0
for _microbit_time in _microbit_main_for_loop(_instructor_code):
    show(counter)
    counter += 1

# Then we have...
# instructor.py
from pedal import *

def grade_loop_iteration(student):
    assert_microbit_displaying(1)
    yield
    assert_microbit_displaying(2)

run_async(grade_loop_iteration)
```

A major advantage over the function approach is that we are no longer beholden to the scope problems of function
definitions. But does this let us achieve our goal?

```python
def run_async():
    student_code = get_program()
    exec(student_code)
```

(Experimented with this in a scratch file. The answer is that it might work for the cpython version, but is incompatible
with the current version of the Skulpt execution model, I *think* (tried it from the student side, but not the instructor).
The basic issue is that exec uses its own separate dictionary that doesn't get updated until after the execution
finishes, I believe, so there's no way to get the "in-progress" student data? Might be solveable.

But the bigger issue with what I was trying there is that the instructor ends up having to stuff all their code in
a function that uses yield quite a bit. I'm not a huge fan of forcing the instructor to think that way?

### Async API?

A more complicated idea is to have blocking API calls for the API, that yield execution back to the instructor.

```python
# answer.py
from microbit.display import show
counter = 0
while True:
    show(counter)
    counter += 1

# instructor.py

run()     # Non-blocking version that executes in a separate thread?
await shown(1)
await shown(2)
```

So that `shown` function is obviously doing some interesting stuff:

```python
def shown(value, timeout=1000):
    '''
    Args:
        timeout (int): How long to wait before giving up on the value being displayed. A message will be shown
            to the students.
    '''
```
"""
from pedal.core.feedback import Feedback
from pedal.core.report import MAIN_REPORT
from pedal.sandbox.commands import get_module

BRIGHTNESS_VALUE = {
    0: 0, '0': 0, ' ': 0, '_': 0, '⬛': 0,
    1: 1, '1': 1,
    2: 2, '2': 2,
    3: 3, '3': 3,
    4: 4, '4': 4,
    5: 5, '5': 5,
    6: 6, '6': 6,
    7: 7, '7': 7,
    8: 8, '8': 8,
    9: 9, '9': 9, '*': 9, '■': 9, '█': 9, '🟥': 9, '⬜': 9, 'X': 9, 'x': 9,
}


class ImageFeedback(Feedback):
    def make_hint(self, differences, report):
        return "\n".join(
            f"{report.format.indent()}{x}, {y}: {report.format.python_expression(repr(actual))}"
            f" instead of {report.format.python_expression(repr(brightness))}"
            for x, y, brightness, actual in differences
        )


class image_not_displayed(ImageFeedback):
    category = Feedback.CATEGORIES.INSTRUCTOR
    title = "Image Not Displayed on Microbit"
    message_template = "The expected image was not displayed on the Microbit.{difference_hint}"

    def __init__(self, expected, max_differences, differences, report=MAIN_REPORT, **kwargs):
        fields = kwargs.setdefault('fields', {})
        fields['expected'] = expected
        fields['max_differences'] = max_differences
        fields['differences'] = differences
        if max_differences:
            fields['difference_hint'] = f"\nThe closest actual image shown had {len(differences)} differences.\n"
            if max_differences is True or len(differences) <= max_differences:
                fields['difference_hint'] += "The differences were at the following positions:\n" + self.make_hint(differences, report)
        else:
            fields['difference_hint'] = ""
        super().__init__(**kwargs)


class image_not_displaying(ImageFeedback):
    category = Feedback.CATEGORIES.INSTRUCTOR
    title = "Image Not Displaying on Microbit"
    message_template = "The expected image is not currently displaying on the Microbit.{difference_hint}"

    def __init__(self, expected, max_differences, differences, report=MAIN_REPORT, **kwargs):
        fields = kwargs.setdefault('fields', {})
        fields['expected'] = expected
        fields['max_differences'] = max_differences
        fields['differences'] = differences
        if max_differences:
            fields['difference_hint'] = f"\nThe image actually shown had {len(differences)} differences.\n"
            if max_differences is True or len(differences) <= max_differences:
                fields['difference_hint'] += "The differences were at the following positions:\n" + self.make_hint(differences, report)
        else:
            fields['difference_hint'] = ""
        super().__init__(**kwargs)


def match_image_dict(target, candidate, **kwargs):
    return False


def match_image_str(target, candidate, **kwargs):
    if ':' in target:
        lines = target.split(':')
    elif '\n' in target:
        lines = target.split('\n')
    else:
        lines = [target[run:run+5] for run in range(0, len(target), 5)]
    differences = []
    for y, (target_line, candidate_line) in enumerate(zip(lines, candidate)):
        for x, (target_b, candidate_b) in enumerate(zip(target_line, candidate_line)):
            brightness = BRIGHTNESS_VALUE.get(target_b, '?')
            if brightness != '?' and brightness != candidate_b:
                differences.append((x, y, brightness, candidate_b))
    return differences


def match_image_2d(target, candidate, **kwargs):
    return False


def match_image(target, candidate, **kwargs):
    if isinstance(target, dict):
        return match_image_dict(target, candidate, **kwargs)
    if isinstance(target, str):
        return match_image_str(target, candidate, **kwargs)
    # Assume 2D sequence
    return match_image_2d(target, candidate, **kwargs)


def assert_microbit_displayed(image, *additional_images, report_differences=5, report=MAIN_REPORT, **kwargs):
    """

    Args:
        image: Either a 2d sequence of integers/strings, a single string of several possible formats, or a dictionary
            mapping positions to values.
        differences: How many differences to report. If `True`, then all pixels that are not equivalent will be
            reported. If an integer, then if the number of differences is less than or equal to that number, they
            will be reported. If `False`, then no differences are shown.
        report:

    Returns:

    """
    microbit = get_module('microbit', report=report)
    smallest_difference = None
    all_images = [image, *additional_images]
    all_history = iter(microbit.display.history)
    for image in all_images:
        for old_image in all_history:
            differences = match_image(image, old_image, **kwargs)
            if not differences:
                break
            if smallest_difference is None or len(smallest_difference) > len(differences):
                smallest_difference = differences
        else:
            return image_not_displayed(image, report_differences, smallest_difference)
    else:
        return False


def assert_microbit_displaying(image, report_differences=5, report=MAIN_REPORT, **kwargs):
    """

    Args:
        image: Either a 2d sequence of integers/strings, a single string of several possible formats, or a dictionary
            mapping positions to values.
        differences: How many differences to report. If `True`, then all pixels that are not equivalent will be
            reported. If an integer, then if the number of differences is less than or equal to that number, they
            will be reported. If `False`, then no differences are shown.
        report:

    Returns:

    """
    microbit = get_module('microbit', report=report)
    differences = match_image(image, microbit.display.current.image, **kwargs)
    if not differences:
        return None
    return image_not_displaying(image, report_differences, differences)


import ast
import dataclasses
import datetime

from pedal.core.feedback import Feedback
from pedal.sandbox.result import is_sandbox_result
from pedal.types import new_types

# acbart: Provide JSONEncoder for Skulpt compatibility
try:
    from json import JSONEncoder
except AttributeError:
    class JSONEncoder:
        pass


class PedalJSONEncoder(JSONEncoder):
    """
    Custom JSON Encoder to handle weird Pedal values nested in Feedback Functions,
    including things like SandboxResult.
    """
    def _iterencode(self, o, markers=None):
        if is_sandbox_result(o):
            return super()._iterencode(o._actual_value, markers)
        else:
            return super()._iterencode(o, markers)

    def iterencode(self, o, _one_shot=False):
        if is_sandbox_result(o):
            return super().iterencode(o._actual_value, _one_shot)
        else:
            return super().iterencode(o, _one_shot)

    def default(self, obj):
        if hasattr(obj, 'to_json'):
            return obj.to_json()
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return (datetime.datetime.min + obj).time().isoformat()
        elif isinstance(obj, new_types.Type):
            return str(obj)
        elif isinstance(obj, ast.AST):
            return obj.__class__.__name__
        try:
            return repr(obj)
        except:
            return '<not serializable>'


def clean_json(obj):
    if is_sandbox_result(obj):
        return clean_json(obj._actual_value)
    elif isinstance(obj, Feedback):
        return clean_json(obj.to_json())
    if isinstance(obj, dict):
        return {clean_json(key): clean_json(value) for key, value in obj.items()}
    elif isinstance(obj, (set, list, tuple)):
        return [clean_json(value) for value in obj]
    else:
        return obj



from pedal.core.final_feedback import FinalFeedback, set_correct_no_errors
from pedal.resolvers.core import make_resolver
from pedal.core.report import MAIN_REPORT
from pedal.core.feedback import Feedback
from pedal.resolvers.simple import by_priority

DEFAULT_CATEGORY_PRIORITY = [
    Feedback.CATEGORIES.INSTRUCTIONS,
    Feedback.CATEGORIES.SYNTAX,
    Feedback.CATEGORIES.MISTAKES,
    Feedback.CATEGORIES.INSTRUCTOR,
    Feedback.CATEGORIES.ALGORITHMIC,
    Feedback.CATEGORIES.RUNTIME,
    Feedback.CATEGORIES.STUDENT,
    Feedback.CATEGORIES.POSITIVE,
    Feedback.CATEGORIES.UNKNOWN
]


def resolve_feedback(feedback):
    """
    This function is used to resolve a single feedback object into a string.

    Args:
        feedback: The feedback object to resolve.

    Returns:
        str: The resolved feedback.
    """
    return "{title}\n{body}".format(title=feedback.title, body=feedback.message)


@make_resolver
def resolve(report=MAIN_REPORT, priority_key=by_priority):
    """
    This function is used to resolve the feedback objects in the report into a FinalFeedback object.

    Args:
        priority_key: The function used to sort the feedback objects.
        report: The report to resolve.

    Returns:
        FinalFeedback: The resolved feedback.
    """

    # Prepare feedbacks
    feedbacks = report.feedback + report.ignored_feedback
    feedbacks.sort(key=priority_key)
    # Create the initial final feedback
    final = set_correct_no_errors(report)
    # Process each feedback in turn
    used = []
    for feedback in feedbacks:
        partial = final.merge(feedback)
        if partial is not None:
            used.append(partial)
    # Override empty message
    final.finalize()
    final.used = used
    report.result = final
    report.resolves.append(final)
    return final



from pedal.assertions.feedbacks import AssertionFeedback
from pedal.core.report import MAIN_REPORT


class missing_function(AssertionFeedback):
    """ Unconditionally asserts that the function is missing. """
    title = "Missing Function"
    message_template = "No function named {name_message} was found."

    def __init__(self, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        super().__init__(fields=fields, **kwargs)


class duplicate_function_definition(AssertionFeedback):
    """ Unconditionally assert that the function is redefined somewhere. """
    title = "Duplicate Function Definition"
    message_template = ("The function {name_message} was defined multiple times, "
                        "on lines {lines_message}.")

    def __init__(self, name, lines, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        lines_message = ", ".join(str(line) for line in lines[:-1])
        lines_message += " and " + str(lines[-1])
        fields['lines'] = lines
        fields['lines_message'] = lines
        super().__init__(fields=fields, **kwargs)


class too_few_parameters(AssertionFeedback):
    """ Too Few Parameters """
    title = "Too Few Parameters"
    message_template = ("The function named {name_message} has fewer "
                        "parameters ({found}) than expected ({expected}).")

    def __init__(self, name, found, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['found'] = found
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class too_many_parameters(AssertionFeedback):
    """ Too Many Parameters """
    title = "Too Many Parameters"
    message_template = ("The function named {name_message} has more "
                        "parameters ({found}) than expected ({expected}).")

    def __init__(self, name, found, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['found'] = found
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class invalid_parameter_type(AssertionFeedback):
    """ Invalid parameter type """
    title = "Invalid Parameter Type"
    message_template = ("The function named {name_message} has a parameter "
                        "named {parameter_message} with an invalid type.")

    def __init__(self, name, parameter, invalid_type, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['parameter'] = parameter
        fields['parameter_message'] = report.format.name(parameter)
        fields['invalid_type'] = invalid_type
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class missing_parameter_type(AssertionFeedback):
    """ Invalid parameter type """
    title = "Missing Parameter Type"
    message_template = ("The function named {name_message} has a parameter "
                        "named {parameter_message}, but that parameter does "
                        "not have a type specified.")

    def __init__(self, name, parameter, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['parameter'] = parameter
        fields['parameter_message'] = report.format.name(parameter)
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class wrong_parameter_type(AssertionFeedback):
    """ Wrong Parameter Type """
    title = "Wrong Parameter Type"
    message_template = ("The function named {name_message} has a parameter "
                        "named {parameter_message} that is {actual_message},"
                        " but should be {expected_message}.")

    def __init__(self, name, parameter, actual, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['parameter'] = parameter
        fields['parameter_message'] = report.format.name(parameter)
        fields['actual'] = actual
        fields['actual_message'] = actual.precise_description()
        fields['expected'] = expected
        fields['expected_message'] = expected.precise_description()
        super().__init__(fields=fields, **kwargs)


class missing_return_type(AssertionFeedback):
    """ Missing Returns Type """
    title = "Missing Return Type"
    message_template = ("The function named {name_message} does not have a "
                        "return type specified in its header.")

    def __init__(self, name, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class invalid_return_type(AssertionFeedback):
    """ Invalid Return Type """
    title = "Invalid Return Type"
    message_template = ("The function named {name_message} has an invalid "
                        "return type in its header.")

    def __init__(self, name, returns, invalid_type, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['returns'] = returns
        fields['invalid_type'] = invalid_type
        fields['expected'] = expected
        super().__init__(fields=fields, **kwargs)


class wrong_return_type(AssertionFeedback):
    """ Wrong Return Type """
    title = "Wrong Return Type"
    message_template = ("The function named {name_message} was expected to "
                        "return {expected_message}, but instead its header "
                        "specifies that it returns {actual_message}. ")

    def __init__(self, name, actual, expected, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        fields['actual'] = actual
        fields['actual_message'] = actual.precise_description()
        fields['expected'] = expected
        fields['expected_message'] = expected.precise_description()
        super().__init__(fields=fields, **kwargs)


class name_is_not_a_function(AssertionFeedback):
    """ The name is not a function """
    title = "Name Is Not a Function"
    message_template = "You defined `{name_message}`, but did not define it as a function."

    def __init__(self, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        super().__init__(fields=fields, **kwargs)


class function_not_available(AssertionFeedback):
    """ The name is not a function """
    title = "Function Not Available"
    message_template = ("You may have defined `{name_message}`, but it was not available"
                        " to be called in the top-level scope. Perhaps you mistakenly defined"
                        " it inside another function or scope?")

    def __init__(self, name, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = kwargs.get("fields", {})
        fields['name'] = name
        fields['name_message'] = report.format.name(name)
        super().__init__(fields=fields, **kwargs)



"""
Resolver that collates all the feedback given into a simple format.
"""
from pedal import Feedback
from pedal.core.final_feedback import FinalFeedback, set_correct_no_errors
from pedal.resolvers.core import make_resolver
from pedal.resolvers.export import PedalJSONEncoder, clean_json
from pedal.resolvers.simple import by_priority
from pedal.core.report import MAIN_REPORT


@make_resolver
def resolve(clean=False, report=MAIN_REPORT):
    """

    Args:
        clean (bool): Whether to clean the JSON output.
        report (Report): The report to resolve. Defaults to MAIN_REPORT.

    Returns:

    """
    feedbacks = report.feedback + report.ignored_feedback
    feedbacks.sort(key=by_priority)
    final = set_correct_no_errors(report)
    result = []
    for feedback in feedbacks:
        final.merge(feedback)
        result.append(feedback.to_json())
    final.finalize()
    report.result = result
    report.resolves.append(result)
    result = {
        'considered': [r for r in result],
        'final': final.to_json()
    }
    if clean:
        result = clean_json(result)
        pedal_json_encoder = PedalJSONEncoder(indent=2, skipkeys=True)
        result = pedal_json_encoder.encode(result)
    return result



from pedal.sandbox.mocked import MockModule
from pedal.utilities.system import IS_PYTHON_36


class MockDesigner(MockModule):
    """
    This will eventually be much more sophisticated. We really need a "MockPygame" that provides a
    window-less version of itself suitable for testing. I think that will be very key for the long-term
    testing behavior of Designer!
    """
    UNKNOWN_FUNCTIONS = []

    SUBMODULES = []

    module_name = 'designer'
    FIELDS = ['Animation', 'Arc', 'COMMON_EVENT_NAMES', 'COMMON_EVENT_NAME_LOOKUP',
              'emoji', '_tifa_definitions',
              'CubicIn', 'CubicInOut', 'CubicOut', 'DEFAULT_WINDOW_TITLE', 'DelayAnimation', 'DesignerObject',
              'Director', 'Event', 'EventHandler', 'GLOBAL_DIRECTOR', 'GameEndException', 'InternalImage',
              'Iterate', 'KNOWN_EVENTS', 'KeyboardKey', 'KeyboardModule', 'Keys', 'Linear', 'LinearTuple',
              'List', 'LiveEventHandler', 'MOUSE_MAP', 'Mods', 'MouseModule', 'MultiAnimation', 'MusicModule',
              'Optional', 'Polar', 'QuadraticIn', 'QuadraticInOut', 'QuadraticOut', 'Random',
              'ReplayEventHandler', 'Request', 'SequentialAnimation', 'SfxModule', 'Sine', 'Union', 'Vec2D',
              'WeakMethod', 'WeakSet', 'Window', 'above',
              'animation', 'arc', 'background_image', 'background_music', 'base64', 'below', 'check_initialized',
              'circle', 'clear_namespace', 'colliding', 'colors', 'continue_music', 'core', 'cursors', 'designer',
              'designer_object', 'destroy', 'difflib', 'disable_keyboard_repeating', 'draw', 'ellipse',
              'enable_keyboard_repeating', 'environ', 'f', 'get_director', 'get_height', 'get_keyboard_delay',
              'get_keyboard_interval', 'get_keyboard_repeat', 'get_mouse_cursor', 'get_mouse_position',
              'get_mouse_visible', 'get_mouse_x', 'get_mouse_y', 'get_music_position', 'get_music_volume',
              'get_positional_event_parameters', 'get_width', 'get_window_color', 'get_window_title',
              'glide_around', 'glide_down', 'glide_in_degrees', 'glide_left', 'glide_right', 'glide_up',
              'group', 'handle', 'helpers', 'hex_to_rgb', 'image', 'imghdr', 'importlib', 'inspect', 'io',
              'is_music_playing', 'json', 'keyboard', 'keys', 'line', 'linear_animation', 'make_suggestions',
              'math', 'mods', 'mouse', 'music', 'objects', 'os', 'path', 'pause', 'pause_music', 'play_music',
              'play_sound', 'positioning', 'pprint', 'pygame', 'queue', 'random', 're', 'rectangle', 'register',
              'rewind_music', 'sequence_animation', 'set_game_state', 'set_keyboard_delay',
              'set_keyboard_interval', 'set_keyboard_repeat', 'set_mouse_cursor', 'set_mouse_position',
              'set_mouse_visible', 'set_music_position', 'set_music_volume', 'set_window_color',
              'set_window_size', 'set_window_state', 'set_window_title', 'set_world_state', 'sfx', 'shape',
              'spin', 'start', 'stop', 'stop_music', 'sys', 'text', 'this_directory', 'unregister', 'urlopen',
              'utilities', 'when',
              'above', 'below', 'beside', 'spin',
              'move_forward', 'move_backward', 'turn_left', 'turn_right', 'go_to', 'go_to_xy', 'go_to_mouse',
              'point_towards', 'point_towards_mouse', 'point_in_direction', 'change_xy', 'change_x',
              'change_y', 'set_x', 'set_y',
              'get_angle', 'get_x', 'get_y',
              'flip_x', 'flip_y', 'set_flip_x', 'set_flip_y', 'set_scale', 'set_scale_x', 'set_scale_y',
              'set_background_image',
              'get_scale', 'get_scale_x', 'get_scale_y', 'get_visible', 'get_flip_x', 'get_flip_y', 'show',
              'hide',
              'grow', 'grow_x', 'grow_y', 'shrink',
              'move_to_x', 'move_to_y', 'move_to', 'move_to_mouse', 'move_to_xy',
              'set_visible', 'change_scale', 'set_emoji_name', 'set_text']

    def __init__(self):
        super().__init__()
        self.calls = []

    def get_submodules(self):
        return {}

    def _reset_designer(self):
        self.calls = []

    def _generate_patches(self):
        if IS_PYTHON_36:
            # Generate patches manually
            return {name: self.getter(name) for name in self.FIELDS}
        else:
            return {'__getattr__': self.getter, '__all__': ['test']}

    def getter(self, key):
        """ If anything asks, we prevent the module. Except for __file__. """
        # Needed to support coverage - it's okay to ask who I am.
        if key == '__file__':
            return self.module_name
        if key == '__all__':
            return self.FIELDS

        def _fake_call_wrapper(*args, **kwargs):
            self.calls.append((key, args, kwargs))
            return self._fake_call(args, kwargs)

        return _fake_call_wrapper

    def _fake_call(self, *args, **kwargs):
        return 0





import unittest
import os
import sys

from pedal.utilities.system import IS_AT_LEAST_PYTHON_311

pedal_library = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, pedal_library)

from tests.execution_helper import Execution, SUCCESS_MESSAGE
from pedal.core.commands import *


class TestFeedback(unittest.TestCase):
    maxDiff = None

    def test_no_errors(self):
        with Execution('a=0\na') as e:
            pass
        self.assertEqual(e.feedback, SUCCESS_MESSAGE)

    def test_success(self):
        with Execution('a=0\na') as e:
            set_success()
        self.assertEqual(e.feedback, "Complete\nGreat work!")

    def test_runtime(self):
        with Execution('0+"A"', old_style_messages=True) as e:
            suppress("analyzer")
        self.assertEqual("""Type Error
I ran your code.

A TypeError occurred:

    Unsupported operand type(s) for +: 'int' and 'str'

The traceback was:
Line 1 of file answer.py
    0+"A"
""" + ('    ^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") +"""
Type errors occur when you use an operator or function on the wrong type of value. For example, using `+` to add to a list (instead of `.append`), or dividing a string by a number.

Suggestion: To fix a type error, you should trace through your code. Make sure each expression has the type you expect it to have.""",
                         e.feedback)

    @unittest.skip
    def test_tifa(self):
        with Execution('0+"A"') as e:
            pass
        self.assertEqual(e.feedback, "TypeError<br><pre>unsupported operand "
                                     "type(s) for +: 'int' and 'str'</pre>\n")


if __name__ == '__main__':
    unittest.main(buffer=False)



"""
A tool for providing dynamic questions to learners.
"""

from pedal.questions.setup import reset, set_seed, _name_hash
from pedal.questions.loader import load_question, SETTING_SHOW_CASE_DETAILS
from pedal.questions.questions import Question
from pedal.questions.pool import Pool
from pedal.questions.graders import FunctionGrader, QuestionGrader

NAME = 'Questions'
SHORT_DESCRIPTION = "Provides dynamic questions to learners"
DESCRIPTION = '''
'''
REQUIRES = []
OPTIONALS = []
CATEGORY = 'Instructions'

__all__ = ['NAME', 'DESCRIPTION', 'SHORT_DESCRIPTION', 'REQUIRES', 'OPTIONALS',
           'Question', 'Pool', 'set_seed', "FunctionGrader", 'QuestionGrader']



"""
Helpful pages:

* Microbit library source code: https://github.com/bbcmicrobit/micropython/tree/master/source/microbit
* Microbit Python API typings: https://github.com/MrYsLab/pseudo-microbit/blob/master/microbit/__init__.py
"""

from pedal.sandbox.mocked import MockModule


def generic_function_capture(self, function_name):
    def _call(*args, **kwargs):
        """ This function does nothing. """
        # TODO: Capture name; does this need to be a decorated function?
        self._unknown_calls.append((function_name, args, kwargs))
    return _call


class MethodExposer:
    def __init__(self):
        self.tracked_functions = {}

    def __call__(self, function):
        self.tracked_functions[function.__name__] = function
        return function


class MockModuleExposing(MockModule):
    expose: MethodExposer
    UNKNOWN_FUNCTIONS: list

    def get_submodules(self):
        return {}

    def _generate_patches(self):
        # Use function descriptor to get a bound version of each method
        fields = {n: f.__get__(self) for n, f in self.expose.tracked_functions.items()}
        fields.update({
            name: generic_function_capture(self, name)
            for name in self.UNKNOWN_FUNCTIONS
        })
        fields.update(self.get_submodules())
        return fields


class MicrobitButton:
    is_pressed: bool
    was_pressed: bool
    presses: int

    def __init__(self):
        self.reset_state()

    def reset_state(self):
        self.is_pressed = False
        self.was_pressed = False
        self.presses = 0


class MicrobitDigitalPin:
    NO_PULL = 0
    PULL_UP = 1
    PULL_DOWN = 2

    value: float
    period: int

    def __init__(self):
        self.reset_state()

    def reset_state(self):
        self.value = 0
        self.period = 35


class MicrobitAnalogDigitalPin(MicrobitDigitalPin):
    pass


class MicrobitTouchPin(MicrobitAnalogDigitalPin):
    touched: bool

    def reset_state(self):
        super().reset_state()
        self.touched = False


class MicrobitDisplay:
    # Actual displayed data (`image_buffer` in c version)
    image: list  #: list[list[int]]
    # Other data
    previous_brightness: int
    active: bool
    strobe_row: int
    #: Histogram of brightnesses
    brightnesses: int
    #: Brightness level => # of pins
    pins_for_brightness: list  #: list[int]

    SIZE = 5

    def __init__(self):
        self.image = [[0 for _ in range(self.SIZE)] for _ in range(self.SIZE)]
        self.reset_state()

    def reset_state(self):
        for x in range(self.SIZE):
            for y in range(self.SIZE):
                self.image[x][y] = 0

    def copy(self):
        return [list(line) for line in self.image]


class MicrobitImage:
    pass


class MicrobitSimulator:
    #: Whether the board is in "panic mode" (error state)
    panic_mode: bool
    #: Status code of the microbit
    status_code: int
    #: Temperature in degrees celsius
    temperature: int
    #: Milliseconds since board started
    time: int

    button_a: MicrobitButton
    button_b: MicrobitButton

    pin0: MicrobitTouchPin
    pin1: MicrobitTouchPin
    pin3: MicrobitTouchPin
    # TODO: Finish these: https://github.com/MrYsLab/pseudo-microbit/blob/master/microbit/__init__.py#L159

    def __init__(self):
        self.button_a = MicrobitButton()
        self.button_b = MicrobitButton()
        self.pin0 = MicrobitTouchPin()
        self.pin1 = MicrobitTouchPin()
        self.pin2 = MicrobitTouchPin()
        self.reset_state()

    def reset_state(self):
        self.panic_mode = False
        self.status_code = 0
        self.temperature = 20
        self.time = 0
        self.button_a.reset_state()
        self.button_b.reset_state()
        self.pin0.reset_state()
        self.pin1.reset_state()
        self.pin2.reset_state()

    def __str__(self):
        if self.panic_mode:
            return "<MicrobitSimulator: PANIC MODE>"
        return "<MicrobitSimulator>"

    def __repr__(self):
        return f"MicrobitSimulator(panic_mode={self.panic_mode})"


class MockMicrobitDisplay(MockModuleExposing):
    """
    If needed, could make the history extremely compact by tracking changes instead, and storing
    them in a single integer. First 6 bits for the X/Y, and then 4 bits for the new value. Could also have some
    special codes, if needed (e.g., for "clear").
    """
    expose = MethodExposer()
    UNKNOWN_FUNCTIONS = []
    SUBMODULES = []
    DEFAULT_LIGHT_LEVEL = 110

    def __init__(self, display):
        super().__init__()
        self.light_level = self.DEFAULT_LIGHT_LEVEL
        self.history = []
        self.current = display
        self.clear()

    @expose
    def clear(self):
        self.light_level = self.DEFAULT_LIGHT_LEVEL
        self.current.reset_state()
        self.record_current()

    @expose
    def show(self, image, delay=400, wait=True, loop=False, clear=False):
        if isinstance(image, MicrobitImage):
            pass
        # TODO: Fix this to use the proper Image api!
        for y, row in enumerate(image):
            for x, brightness in enumerate(row):
                self.current.image[y][x] = brightness
        self.record_current()

    @expose
    def set_pixel(self, x, y, value):
        self.current.image[y][x] = value
        self.record_current()

    @expose
    def get_pixel(self, x, y):
        return self.current.image[y][x]

    def record_current(self):
        self.history.append(self.current.copy())

    @expose
    def read_light_level(self):
        return self.light_level

    @expose
    def set_light_level(self, light_level):
        self.light_level = light_level


class MockMicrobit(MockModuleExposing):
    """
    Mock Microbit library that can be used to capture data from the students' program execution.
    """
    expose = MethodExposer()
    UNKNOWN_FUNCTIONS = []

    SUBMODULES = [
        'microbit.display'
    ]

    def __init__(self):
        super().__init__()
        self.state = MicrobitSimulator()
        self.display = MockMicrobitDisplay(MicrobitDisplay())
        self._unknown_calls = []

    def __repr__(self):
        return f"MockMicrobit(_state={self._state!r})"

    def __str__(self):
        return f"<MockMicrobit ({self._state!r})>"


    @expose
    def panic(self, n: int = 999):
        self.state.panic_mode = True
        self.state.status_code = n

    @expose
    def reset(self):
        self.state.reset_state()

    @expose
    def sleep(self, n: float):
        self.state.time += n

    @expose
    def running_time(self):
        return self.state.time

    def get_submodules(self):
        return {
            'display': self.display
        }





"""
Jupyter Environment.

Our local context was an "Introduction to Computational Thinking" course. We incorporated a special
toolbar that hid some features and exposed a new "run all" button that ran the students code and then
also a special cell at the end which grabbed the students' code.

Adds in two special magic commands when integrated into a Jupyter environment:
    %grade
    %grade_blockpy

If you want to use this, we expect there to be some work involved. We haven't really made it portable
and extensible the way we have everything else, because we aren't sure if anyone would want to use it.
We really need someone with a dedicated interest to shepard it.

"""

# Built-in imports
import json
import requests

# IPython imports
from IPython.core.magic import (Magics, magics_class, line_magic, cell_magic, line_cell_magic)
from IPython.display import Javascript, display
from IPython.utils.io import capture_output, CapturedIO

# Logging imports
import os
import sys
from warnings import warn
# from traitlets import Bool
import time

# TODO: Opportunity here to add in requests-cache. This would allow us to avoid
# the repeated trip. However, you'll need to handle expiring the cache in a
# smart way. One option is to write a command line script to just wipe as
# necessary. Simply deleting the cache file would be pretty easy, assuming it
# installs per user.

# TODO: MAKE THIS A CONFIG SETTING
BLOCKPY_URL = 'https://think.cs.vt.edu/blockpy/blockpy/load_assignment_give_feedback'


def get_response_error(response):
    """
    Transform a Response object into a friendlier string.

    Args:
        response (requests.Response): A Requests reponse object to parse for
                                      some kind of error.
    Returns:
        str: A string representation of the URL response.
    """
    return "{} {}: {}".format(response.status_code, response.reason,
                              response.text)


def download_on_run(assignment_id):
    """
    Download the on_run (give_feedback) code to use to test their solution.

    Args:
        assignment_id (int OR str): The ID of the assignment to get the
                                    on_run code for.
    Returns:
        bool: Whether or not the request was successful.
        str: If unsuccesful, a message to display to the user. Otherwise, it'll
             be the on_run code.
    """
    data = {'assignment_id': assignment_id}
    try:
        response = requests.get(BLOCKPY_URL, data=data)
    except Exception as error:
        return False, str(error)
    try:
        result = response.json()
    except ValueError:
        # Failed to parse the JSON; perhaps it was some text data?
        return False, get_response_error(response)
    if result['success']:
        return True, result['give_feedback']
    else:
        return False, result['message']


PEDAL_PIPELINE = '''
from pedal.core.report import MAIN_REPORT
MAIN_REPORT.clear()

# Load in some commonly used tools
from pedal.cait.cait_api import parse_program
from pedal.sandbox.commands import *
from pedal.core.commands import *

from pedal.environments.blockpy import setup_environment
# Initialize the BlockPy environment
pedal = setup_environment(skip_tifa=False,
                          skip_run=False,
                          inputs={inputs},
                          main_file='answer.py',
                          main_code={student_code})
student = pedal.fields['student']

{on_run}

from pedal.resolvers.simple import resolve
final = resolve()
SUCCESS = final.success
SCORE = final.score
CATEGORY = final.category
LABEL = final.title
MESSAGE = final.message
DATA = final.data
HIDE = final.hide_correctness
'''


def blockpy_grade(assignment_id, student_code, inputs):
    """
    Helper function to capture the request from the server.

    Args:
        assignment_id (int): The assignment ID to look up and use the on_run
                             code for.
        student_code (str): The code that was written by the student.

        inputs (str): The inputs to queue into the assignment

    Returns:
        str: The HTML formatted feedback for the student.
    """
    successful_download, on_run = download_on_run(assignment_id)
    # If it failed, let's display some information about why.
    if not successful_download:
        return on_run
    return execute_on_run_code(on_run, student_code, inputs)


def execute_on_run_code(on_run, student_code, inputs):
    """
    Actually execute the on_run code for the given student code.
    """
    # Even though the student code is a string, we need to escape it to prevent
    # any weirdness from being in the instructor code.
    escaped_student_code = json.dumps(student_code)
    instructor_code = PEDAL_PIPELINE.format(on_run=on_run,
                                            student_code=escaped_student_code,
                                            # inputs=','.join(inputs))
                                            inputs=inputs)
    # Execute the instructor code in a new environment
    global_variables = globals()
    compiled_code = compile(instructor_code, 'instructor_code.py', 'exec')
    exec(compiled_code, global_variables)
    category = global_variables['CATEGORY']
    label = global_variables['LABEL']
    message = global_variables['MESSAGE']
    # In some cases, we might want to override how the text is rendered.
    if category.lower() == 'instructor' and label.lower() == 'explain':
        category = "Instructor Feedback"
        label = ''
    # Return the result as HTML
    return '''<strong>{}</strong>: {}<br>{}'''.format(category, label, message)


# The following string literals are used to create the JavaScript code that
# creates the Python code that will execute the instructor's feedback code
# using the student's Python code.

# Extract out the student code, embed the result
EXTRACT_STUDENT_CODE = r"""
// Convert Notebook cells to a string of Python code
var makePython = function(cell) {
    if (cell.cell_type == "code") {
        // Code is embedded unchanged, unless it is magic
        var source = cell.get_text();
        if (source.startsWith('%')) {
            // Skip magic
            return '';
        } else {
            return source;
        }
    } else if (cell.cell_type == "markdown" ||
               cell.cell_type == "raw") {
        // Markdown and text is wrapped in a string.
        var escaped_text = cell.get_text().replace(/'''/g, "\\'\\'\\'");
        return "'''"+escaped_text+"'''";
    }
}
var isUsable = function(cell) {
    return cell.cell_type == "code" ||
           cell.cell_type == "markdown" ||
           cell.cell_type == "raw";
}
var cells = Jupyter.notebook.get_cells();
var source_code = cells.filter(isUsable).map(makePython).join("\n");
source_code = JSON.stringify(source_code);
console.log(source_code);
// Start constructing the feedback code (which will be Python).
var on_run_code = [];
on_run_code.push("student_code="+source_code);
"""

# Retrieve the last cell, and also recolor it a little for style
ANIMATE_LAST_CELL = r"""
// While we are accessing the server, recolor the last cell a little.
var last = null;
if (cells.length > 0) {
    last = cells[cells.length-1];
    $(last.element).animate({"background-color": "#E0E6FF"}, 1000);
}
"""

# If the %grade magic is used, we run the code directly.
LOCAL_GRADE = r'''
on_run_code.push("from pedal.environments.jupyter import execute_on_run_code");
on_run_code.push('print(execute_on_run_code({on_run_code}, student_code, {inputs}))');
'''

# If the %grade_blockpy magic is used, we need to get the on_run from blockpy.
BLOCKPY_GRADE = r'''
on_run_code.push("from pedal.environments.jupyter import blockpy_grade");
on_run_code.push('import json')
on_run_code.push('inputs = {inputs}')
console.log('inputs = {inputs}')
on_run_code.push("print(blockpy_grade({assignment}, student_code, inputs))");
'''

# This chunk actually performs the on_run code execution using the kernel.
EXECUTE_CODE = r'''
on_run_code = on_run_code.join("\n");
console.log(on_run_code);
var kernel = IPython.notebook.kernel;
if (kernel !== null) {
    var t = kernel.execute(on_run_code, { 'iopub' : {'output' : function(x) {
        if (x.msg_type == "error") {
            // If this was an error, show the traceback properly.
            if (last !== null) {
                last.output_area.append_error(x.content);
                console.error(x);
            } else {
                console.error("Could not append to final cell.", x);
            }
        } else if (!x.content.data && x.content.text) {
            // If it was valid data, we show it as HTML.
            console.log(x);
            element.html(x.content.text.replace(/\n/g, "<br>"));
        } else {
            // I'm not sure what it is - better dump it on the console.
            console.log(x);
        }
        // Decolor the last cell if it was there.
        if (last !== null) {
            last = cells[cells.length-1];
            $(last.element).animate({"background-color": "white"}, 1000);
        }
    }}});
}'''


@magics_class
class GradeMagic(Magics):
    """
    This class holds the magic for the %grade and %grade_blockpy
    """

    @line_magic
    def grade_logstart(self, line=""):
        """

        Args:
            line:
        """
        # ######Logging
        ts = time.time()
        logger = self.shell.logger  # logging
        old_logfile = self.shell.logfile  # logging
        directory = os.path.expanduser("log_folder{}~/".format(line))
        logfname = os.path.expanduser("log_folder{}~/log_{}.py~".format(line, ts))
        self.shell.logfile = logfname
        loghead = u'# IPython log file\n\n'
        try:
            os.makedirs(directory, exist_ok=True)
            logger.logstart(logfname, loghead, 'rotate', True, True,
                            True)
        except BaseException:
            self.shell.logfile = old_logfile
            warn("Couldn't start log: %s" % sys.exc_info()[1])
        self.shell.run_code("input = __builtins__.input")
        self.shell.run_code("print = __builtins__.print")
        self.shell.run_code("sum = __builtins__.sum")
        self.shell.run_code("len = __builtins__.len")

    @line_magic
    def grade_logstop(self, line=""):
        """

        Args:
            line:
        """
        self.shell.logger.logstop()

    def logging(self):
        """

        """
        # ######Logging
        ts = time.time()
        logger = self.shell.logger  # logging
        old_logfile = self.shell.logfile  # logging
        """
        import subprocess
        p = subprocess.Popen("uname -n", stdout=subprocess.PIPE, shell=True)
        (output, err) = p.communicate()  # output contains the username
        """
        logfname = os.path.expanduser("log_folder~/log_{}.py~".format(ts))
        self.shell.logfile = logfname
        loghead = u'# IPython log file\n\n'
        try:
            logger.logstart(logfname, loghead, 'rotate', False, True,
                            True)
        except BaseException:
            self.shell.logfile = old_logfile
            warn("Couldn't start log: %s" % sys.exc_info()[1])
        logger.timestamp = False
        input_hist = self.shell.history_manager.input_hist_raw
        logger.log_write(u'\n'.join(input_hist[1:]))
        logger.log_write(u'\n')
        logger.timestamp = True
        self.shell.logger.logstop()
        # ######Logging

    # noinspection PyMethodMayBeStatic
    def grade_parser(self, line, cell=None):
        """

        Args:
            line:
            cell:

        Returns:

        """
        if ',' in line:
            if cell is None:
                assignment, line = line.split(",", maxsplit=1)
            else:
                assignment = None
            inputs = json.dumps(line.split(",")[1:])
            #inputs = "\\'" + inputs[1:len(inputs) - 1] + "\\'"
        else:
            if cell is None:
                assignment, inputs = line, ""
            else:
                inputs = line
                assignment = ""
            inputs = json.dumps(inputs.split(",")[1:])
        return {"inputs": inputs, "assignment": assignment}

    # noinspection PyMethodMayBeStatic
    def unified_helper(self, local_code, **kwargs):
        """

        Args:
            local_code:
            **kwargs:

        Returns:

        """
        code = EXTRACT_STUDENT_CODE
        code += ANIMATE_LAST_CELL
        code += local_code.format(**kwargs)
        code += EXECUTE_CODE
        return code

    @cell_magic
    def grade(self, line="", cell=""):
        """

        Args:
            line:
            cell:

        Returns:

        """
        dump = self.grade_parser(line, cell)
        code = self.unified_helper(LOCAL_GRADE, on_run_code="INSTRUCTOR_CODE", inputs=dump['inputs'])
        cell = cell.replace("\\", "\\\\")
        cell = cell.replace("\n", "\\n")
        cell = cell.replace("'", "\\'")
        cell = cell.replace('"', '\\"')
        # Runs this code in the kernel as python code
        # Can also run compiled code
        self.shell.run_code("INSTRUCTOR_CODE = " + '"' + cell + '"')
        # TODO: This was the easier way for me to get this to work
        #  This might be worth using in more depth to have less translation
        #  to and from javascript. See usage_examples
        return display(Javascript(code))

    @line_cell_magic
    def usage_examples(self, line="", cell="print('running cell')\nprint('running cell2')"):
        """

        Args:
            line:
            cell:
        """
        # Runs code in the kernel's context
        self.shell.run_code("print('fun')")

        # Runs code in kernel's context using compiled code
        sample = compile(cell, "usage_examples.py", "exec")
        self.shell.run_code(sample)

        # runs javascript code
        self.shell.run_cell_magic("javascript", "", "console.log('I do JAVASCRIPT');\n")
        # Maybe can use javascript execution to pass things around...not sure though...can't get it to work
        # You can pass values, but it doesn't seem to work unless you run it again.
        # https://michhar.github.io/javascript-and-python-have-a-party/

        self.shell.run_cell_magic(
            "javascript", "",
            # js_code = Javascript(
            """var callbacks = { iopub : { output: function(out_data){ console.log(out_data) } } };\n"""
            """var code = "fun = 12";\n"""
            """IPython.notebook.kernel.execute(code);\n""")
        # handle = display(js_code, display_id="usage_examples")
        # handle.update(handle)
        self.shell.run_cell_magic("javascript", "", "console.log('I do JAVASCRIPT TOO!!');\n")
        # captures standard output, standard error, etc. and stops or not stops it
        # class IPython.utils.capture.capture_output(stdout=True, stderr=True, display=True)
        # Note that Tracebacks aren't put in standard error?
        with capture_output(True, False, False) as captured:
            print(dir(self))
            self.shell.run_code("print(fun)")
            sys.stderr.write("spam\n")
        print("I captured stdout")
        print(captured.stdout)
        print("I captured stderr")
        print(captured.stderr)

    @line_magic
    def grade_blockpy(self, line=""):
        """

        Args:
            line:

        Returns:

        """
        dump = self.grade_parser(line)
        code = self.unified_helper(BLOCKPY_GRADE, assignment=dump["assignment"], inputs=dump["inputs"])
        return display(Javascript(code))


def load_ipython_extension(ipython):
    """
    Register this plugin with Jupyter Notebooks. Although it is allegedly
    necessary in order to make this a plugin, we do not actually use it.
    """
    ipython.register_magics(GradeMagic)



import os
import sys
import io
from contextlib import redirect_stdout
import unittest

from pedal import MAIN_REPORT, call
from pedal.utilities.system import IS_AT_LEAST_PYTHON_311

pedal_library = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, pedal_library)

here = "" if os.path.basename(os.getcwd()) == "tests" else "tests/"

from tests.execution_helper import ExecutionTestCase
from pedal.environments.terminal import setup_environment
from pedal.source import next_section, verify_section, check_section_exists, set_source_file, separate_into_sections
from pedal.core.commands import compliment, clear_report, contextualize_report, log, explain
from pedal.assertions import assert_equal, prevent_ast


class TestEnvironmentTerminal(ExecutionTestCase):

    def checkFeedback(self, actual, expected):
        self.assertEqual(expected.strip(), actual.strip())

    def test_correct(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="""1+2""")
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
[;7mYOUR CODE[0;0m We ran your code. Here's the output:


[;7mFEEDBACK[0;0m Based on your code, here are some tips and recommendations:

[1;92m✓[0;0m️  Your code ran successfully.

Great work! Based on your code submitted, there are no other recommendations available. You can proceed to the next page by using the “Next” button in your lesson.
''')

    def test_empty_source(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="")
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
[;7mYOUR CODE[0;0m We ran your code. Here's the output:


[;7mFEEDBACK[0;0m Based on your code, here are some tips and recommendations:

[1;91m✗[0;0m  No Source Code

Source code file is blank.
''')

    def test_syntax_error(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="""1 1 2 3""")
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
[;7mYOUR CODE[0;0m We ran your code. Here's the output:


[;7mFEEDBACK[0;0m Based on your code, here are some tips and recommendations:

[1;91m✗[0;0m  Syntax Error

Bad syntax on line [4m1[0;0m.
Suggestion: Check line [4m1[0;0m, the line before it, and the line after it. Ignore blank lines.

[;7mTERMINAL OUTPUT[0;0m For reference, here is the original Python error:

Line [4m1[0;0m of file [4manswer.py[0;0m
    1 1 2 3
''' + ('      ^\n' if IS_AT_LEAST_PYTHON_311 else "") + '''
SyntaxError: Invalid syntax.
''')

    def test_tifa_feedback(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="""1+''""")
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
[;7mYOUR CODE[0;0m We ran your code. Here's the output:


[;7mFEEDBACK[0;0m Based on your code, here are some tips and recommendations:

[1;91m✗[0;0m  Incompatible types

You used an addition operation with an integer and a string on line [4m1[0;0m. But you can't do that with that operator. Make sure both sides of the operator are the right type.

[;7mTERMINAL OUTPUT[0;0m For reference, here is the original Python error:

Line [4m1[0;0m of file [4manswer.py[0;0m
    1+''
''' + ('    ^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + '''
TypeError: Unsupported operand type(s) for +: 'int' and 'str'
    ''')

    def test_runtime_feedback(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="""import json\njson.loads('1')+''""")
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
[;7mYOUR CODE[0;0m We ran your code. Here's the output:


[;7mFEEDBACK[0;0m Based on your code, here are some tips and recommendations:

[1;91m✗[0;0m  Type Error

I ran your code.

A TypeError occurred:

    Unsupported operand type(s) for +: 'int' and 'str'

Type errors occur when you use an operator or function on the wrong type of value. For example, using `+` to add to a list (instead of `.append`), or dividing a string by a number.

Suggestion: To fix a type error, you should trace through your code. Make sure each expression has the type you expect it to have.

[;7mTERMINAL OUTPUT[0;0m For reference, here is the original Python error:

Line [4m2[0;0m of file [4manswer.py[0;0m
    json.loads('1')+''
''' + ('    ^^^^^^^^^^^^^^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + '''
TypeError: Unsupported operand type(s) for +: 'int' and 'str'
''')

    def test_explain(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="""import json\njson.loads('1')+''""")
            explain("You should not be using the JSON library at all.")
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
[;7mYOUR CODE[0;0m We ran your code. Here's the output:


[;7mFEEDBACK[0;0m Based on your code, here are some tips and recommendations:

[1;91m✗[0;0m  Instructor Feedback

You should not be using the JSON library at all.

[;7mTERMINAL OUTPUT[0;0m For reference, here is the original Python error:

Line [4m2[0;0m of file [4manswer.py[0;0m
    json.loads('1')+''
''' + ('    ^^^^^^^^^^^^^^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + '''
TypeError: Unsupported operand type(s) for +: 'int' and 'str'
''')


    def test_call_assertion(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="""def x(): return 1+''\nx""")
            assert_equal(call('x'), 5)
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
[;7mYOUR CODE[0;0m We ran your code. Here's the output:


[;7mFEEDBACK[0;0m Based on your code, here are some tips and recommendations:

[1;91m✗[0;0m  Type Error

I ran the code:
    x()

A TypeError occurred:

    Unsupported operand type(s) for +: 'int' and 'str'

Type errors occur when you use an operator or function on the wrong type of value. For example, using `+` to add to a list (instead of `.append`), or dividing a string by a number.

Suggestion: To fix a type error, you should trace through your code. Make sure each expression has the type you expect it to have.

[;7mTERMINAL OUTPUT[0;0m For reference, here is the original Python error:

Line [4m1[0;0m of file [4manswer.py[0;0m in x
    def x(): return 1+''
''' + ('                    ^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + '''
TypeError: Unsupported operand type(s) for +: 'int' and 'str'
''')


import unittest
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pedal.core.commands import *
from pedal.tifa import tifa_analysis
from pedal.core.formatting import Formatter
from pedal import MAIN_REPORT, Submission, run
from pedal.resolvers import simple


class CustomFormatter(Formatter):
    def name(self, value):
        return f"<code>{value}</code>"

    def line(self, line_number):
        return f"<a href=''>{line_number}</a>"


class TestCode(unittest.TestCase):

    def test_custom_formatter(self):
        clear_report()
        contextualize_report(Submission(main_code="alpha=0"))
        set_formatter(CustomFormatter)
        tifa_analysis()

        #if get_all_feedback():
        #    print(get_all_feedback()[0].message)

        final = simple.resolve()
        self.assertEqual(final.message,
                         "The variable <code>alpha</code> was given a value on line <a href=''>1</a>, but was never used after that.")


class TestMessageTemplateInCore(unittest.TestCase):

    def test_gently_fields(self):
        clear_report()

        contextualize_report(Submission(main_code="alpha=0"))
        set_formatter(CustomFormatter)
        gently(message_template="This {where:line} should get formatted.",
               fields={"where": 27})

        final = simple.resolve()
        self.assertEqual(final.message, "This <a href=''>27</a> should get formatted.")

    def test_gently_fields_implicit(self):
        clear_report()

        contextualize_report(Submission(main_code="alpha=0"))
        set_formatter(CustomFormatter)
        gently(message_template="This {where:line} should get formatted.",
               where=27)

        final = simple.resolve()
        self.assertEqual(final.message, "This <a href=''>27</a> should get formatted.")


if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Standardized Report format for the result of running Pedal CLI's Stats mode.


Need to capture
    Each submission
        Errors
        Stdout?
        Feedback Functions generated
            All the usual fields
        Final Feedback chosen
    The ICS used on it
    Any metadata surrounding that
        User/Assignment/Course

I'm thinking this has a Dataframe-esque interface, where you can do things like StatReport.scripts to get all the scripts.

list of:
  script (str): The instructor code
  submission:
    user:
      email (str): User's email address
      ...
    assignment:
      name (str): The name of this assignment
    ...
    execution:
      timestamp (str)
      event_id (int)
    files (dict[str: str]): map of filenames to contents
  environment (str): what default configuration was used
  result:
    output (str): Any text piped to stdout by the instructor's code
    error (Exception): Any errors triggered by the instructor's code
    resolution:
      final:
        success (bool): Whether the final feedback was correct
        score (float): Any partial credit assigned
        category (str): The final Feedback Category chosen
        title (str): The final feedback's title
        label (str): The final feedback's label
        ...
      considered (list of): [
        correct (bool): Whether this particular feedback was correct
        active (bool): Whether this feedback was detected
        label (str): The label of this feedback
        ...
      ]
"""

def flatten_nested_json(key, data):
    result = {}
    if isinstance(data, dict):
        for subkey, value in data.items():
            flattened = flatten_nested_json(subkey, value)
            for flat_key, flat_value in flattened.items():
                result[f"{key}.{flat_key}"] = flat_value
    else:
        result[key] = data
    return result


class StatReport:
    script: str
    environment: str

    submission: dict

    result: dict

    def __init__(self, raw):
        self._raw = raw

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._raw[item]
        if isinstance(item, str):
            return [r[item] for r in self._raw]

    def get_all_feedback(self, with_submissions=True):
        data = [dict(**e) if not with_submissions else
                dict(**e, **flatten_nested_json('submission', r['submission']))
                for r in self._raw
                for e in r['result']['considered']]
        return data

    def get_final_feedback(self, with_submissions=True):
        data = [dict(**r['result']['final']) if not with_submissions else
                dict(**r['result']['final'],
                     **flatten_nested_json('submission', r['submission']))
                for r in self._raw]
        return data

    def get_errors(self):
        return [r['result']['error'] for r in self._raw if r['result']['error']]

    def get_outputs(self):
        return [r['result']['output'] for r in self._raw if r['result']['output']]

    def __getattr__(self, attr):
        if attr == 'errors':
            return self.get_errors()
        elif attr == 'outputs':
            return self.get_outputs()
        raise AttributeError("%r object has no attribute %r" %
                             (self.__class__.__name__, attr))



"""
Mocked functions that can be used to prevent malicious or accidental `eval`
behavior.
"""
import re
import sys
import types
from io import StringIO

from pedal.core.report import Report
from pedal.sandbox.exceptions import (SandboxNoMoreInputsException,
                                      SandboxPreventModule)
from pedal.utilities.system import IS_PYTHON_36


def do_nothing(*args, **kwargs):
    """ A function that does absolutely nothing. """


def _disabled_compile(source, filename, mode, flags=0, dont_inherit=False):
    """
    A version of the built-in `compile` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'compile'.")


def _disabled_eval(object, globals=globals(), locals=None):
    """
    A version of the built-in `eval` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'eval'.")


# -------------------------------------------------------------


def _disabled_exec(object, globals=globals(), locals=None):
    """
    A version of the built-in `exec` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'exec'.")


# -------------------------------------------------------------


def _disabled_globals():
    """
    A version of the built-in `globals` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'globals'.")


class FunctionNotAllowed(Exception):
    """ Exception created when a function that is not allowed is used. """


def disabled_builtin(name):
    """

    Args:
        name:

    Returns:

    """
    def _disabled_version(*args, **kwargs):
        raise FunctionNotAllowed("You are not allowed to call '{}'.".format(name))
    return _disabled_version


# TODO: This needs to be a field in the Sandbox object's data
_OPEN_FORBIDDEN_NAMES = re.compile(r"(^[./])|(\.py$)")
_OPEN_FORBIDDEN_MODES = re.compile(r"[wa+]")

# TODO: Allow the user to give a function instead of a REGEX
# TODO: We need to mock the whole OS library, honestly, not just `open`
def create_open_function(report: Report, forbidden_names=_OPEN_FORBIDDEN_NAMES, forbidden_modes=_OPEN_FORBIDDEN_MODES):
    """
    Creates a new version of the `open` built-in that will avoid looking at certain filenames,
    and will also respect the given Report's Submission's Files.

    Args:
        report:

    Returns:

    """
    def _restricted_open(name, mode='r', buffering=-1):
        if forbidden_names.search(name):
            raise RuntimeError("The filename you passed to 'open' is restricted.")
        elif forbidden_modes.search(mode):
            raise RuntimeError("You are not allowed to 'open' files for writing.")
        elif report.submission and name in report.submission.files:
            return StringIO(report.submission.files[name])
        else:
            return ORIGINAL_BUILTINS['open'](name, mode, buffering)
    return _restricted_open

# TODO: Allow this to be flexible


def create_import_function(report: Report, sandbox):
    """
    Creates a new version of the `__import__` function (which is used by the import keyword)
    that will not let students import Pedal or its submodules.

    Args:
        report:

    Returns:

    """
    def _restricted_import(module_name, globals=None, locals=None, fromlist=(), level=0):
        filename = module_name.replace(".", "/")+".py"
        if module_name == 'pedal' or module_name.startswith('pedal.'):
            raise RuntimeError("You cannot import pedal!")
        elif report.submission and filename in report.submission.files:
            if module_name not in sys.modules:
                contents = report.submission.files[filename]
                return sandbox._import(contents, module_name, filename, sandbox.threaded)
        return ORIGINAL_BUILTINS['__import__'](module_name, globals, locals, fromlist, level)
    return _restricted_import


try:
    __builtins__
except NameError:
    _default_builtins = {'globals': globals,
                         'locals': locals,
                         'open': open,
                         'input': input,
                         '__import__': __import__}
else:
    if isinstance(__builtins__, types.ModuleType):
        _default_builtins = __builtins__.__dict__
    else:
        _default_builtins = __builtins__

ORIGINAL_BUILTINS = {
    'globals': _default_builtins['globals'],
    'locals': _default_builtins['locals'],
    'open': _default_builtins['open'],
    'input': _default_builtins['input'],
    'print': _default_builtins.get('print'),
    'exec': _default_builtins.get('exec', _disabled_exec),
    'eval': _default_builtins.get('eval', _disabled_eval),
    'compile': _default_builtins.get('compile', _disabled_compile),
    'exit': disabled_builtin('exit'),
    '__import__': _default_builtins['__import__']
}


def make_inputs(input_list, repeat=None):
    """
    Helper function for creating mock user input.

    Params:
        input_list (list of str): The list of inputs to be returned
    Returns:
        function (str=>str): The mock input function that is returned, which
                             will return the next element of input_list each
                             time it is called.
    """
    generator = iter(input_list)

    def mock_input(prompt=''):
        """

        Args:
            prompt:

        Returns:

        """
        print(prompt)
        try:
            return next(generator)
        except StopIteration as SI:
            if repeat is None:
                # TODO: Make this a custom exception
                raise SandboxNoMoreInputsException("User had no more input to give.")
            else:
                return repeat

    return mock_input


class PrintingStringIO(StringIO):
    _ORIGINAL_STDOUT = sys.stdout

    def __init__(self, stdout=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._original_stdout = self._ORIGINAL_STDOUT if stdout is None else stdout

    def write(self, s):
        self._original_stdout.write(s)
        return super().write(s)

    def flush(self):
        self._original_stdout.flush()
        return super().flush()

    def writelines(self, lines):
        self._original_stdout.writelines(lines)
        return super().writelines(lines)


def make_fake_output(also_print=False) -> StringIO:
    """
    Creates a fake output stream that can be used to capture the output
    of a function. This is useful for testing functions that print things.

    Args:
        also_print (bool): Whether to also print the output to the console.

    Returns:
        StringIO: The fake output stream.
    """
    fake_output = StringIO()


    return fake_output


def create_module(module_names):
    """
    Creates empty ModuleTypes based on the ``module_name``. Correctly
    creates parent modules for submodules as needed to fill in the path.

    Args:
        module_name: A dot-separated string of modules, as if for ``import``. A list of
            names can also be passed in instead. If multiple names are given, they should all have the same root!

    Returns:
        :py:class:`types.ModuleType`: The root module created.
        dict[str, :py:class:`types.ModuleType`]: A dictionary of newly created
            modules, including the root and the actual target.
        :py:class:`types.ModuleType`: The target module created.
    """
    if isinstance(module_names, str):
        module_names = [module_names]
    if not module_names:
        raise ValueError("No modules provided; need at least one module to create!")
    root = None
    modules = {}
    for module_name in module_names:
        submodule_names = module_name.split(".")
        if root is None:
            root = types.ModuleType(submodule_names[0])
        modules[submodule_names[0]] = root
        reconstructed_path = submodule_names[0]
        for submodule_name in submodule_names[1:]:
            reconstructed_path += "." + submodule_name
            new_submodule = types.ModuleType(reconstructed_path)
            setattr(root, submodule_name, new_submodule)
            modules[reconstructed_path] = new_submodule
    return root, modules, modules[module_names[0]]


class MockModule:
    SUBMODULES = {}

    def _generate_patches(self):
        return {k: v for k, v in vars(self).items()
                if not k.startswith('_')}

    def add_to_module(self, main_module, submodules=None):
        for name, value in self._generate_patches().items():
            setattr(main_module, name, value)
        for submodule_name in self.SUBMODULES:
            property_name = submodule_name.split('.', maxsplit=1)[1]
            mock_submodule = getattr(main_module, property_name)
            # the actual created Module object
            submodule = submodules[submodule_name]
            # Populate the actual module object with the fake submodule's data
            mock_submodule.add_to_module(submodule)
            # Update the main module to point to the actual module, instead of the fake submodule
            setattr(main_module, property_name, submodule)


class MockDictModule(MockModule):
    def __init__(self, fields):
        self.fields = fields

    def _generate_patches(self):
        return self.fields


class BlockedModule(MockModule):
    def __init__(self, name):
        self.module_name = name

    def _generate_patches(self):
        return {'__getattr__': self.getter}

    def getter(self, key):
        """ If anything asks, we prevent the module. Except for __file__. """
        # Needed to support coverage - it's okay to ask who I am.
        if key == '__file__':
            return self.module_name
        else:
            self.prevent_module()

    def prevent_module(self, *args, **kwargs):
        """

        Args:
            **kwargs:
        """
        raise SandboxPreventModule(f"You cannot import `{self.module_name}` from student code.")


class MockPedal(BlockedModule):
    """ TODO: Deprecated? """
    module_name = "pedal"


def generic_function_capture(self, function_name):
    def _call(*args, **kwargs):
        """ This function does nothing. """
        # TODO: Capture name; does this need to be a decorated function?
        self._unknown_calls.append((function_name, args, kwargs))
    return _call


class MethodExposer:
    def __init__(self):
        self.tracked_functions = {}

    def __call__(self, function):
        self.tracked_functions[function.__name__] = function
        return function

    def add_with_name(self, function, name):
        self.tracked_functions[name] = function
        return function


class MockModuleExposing(MockModule):
    expose: MethodExposer
    UNKNOWN_FUNCTIONS: list

    def get_submodules(self):
        return {}

    def _generate_patches(self):
        # Use function descriptor to get a bound version of each method
        fields = {n: f.__get__(self) for n, f in self.expose.tracked_functions.items()}
        fields.update({
            name: generic_function_capture(self, name)
            for name in self.UNKNOWN_FUNCTIONS
        })
        fields.update(self.get_submodules())
        return fields


from pedal.sandbox.library import *



"""
A tool for running Pedal from the Command Line

./pedal command line interface
    simple example (single ICS, single student file)
    simple sandbox for just quickly running students code with no frills or extra behavior
        Maybe this is the "default" ICS?
    unit test a given curriculum (many ICS, many student files per)
    regrade some assignments, spit out subject/assignment/context (one ICS, many many student files)
    get summary stats from ProgSnap file (many ICS, many submissions
    Spit out the justifications alongside the potential feedback`
    Run the students code in a sandbox and spit out the runtime results

ProgSnap file with submissions AND assignments (+other info)
    Just one parameter then
ProgSnap file with only submissions, ICS is separate
    Two parameters
Single ICS file
    Single student file
    Archive Format
        Regular directory
        ZIP
        ProgSnap Folder
        ProgSnap SQL
    Structure
        Each file is a python file representing their submission:
            .+\.py
            Could have accompanying Markdown file
        Each folder is a submission
            .+/.+\.py
Folder of ICS files
    Potentially, submissions are within an accompanying "submissions" folder
    Regex for matching ICS files?



argparse options
    -i <instructor_control_script | python file or directory of scripts>
    -s <student_submission | single python file, directory of files, progsnap file>
    -o <output_file>
    -f <output format: Feedback, Grade, Debug mode, Unit Tests, research stats, Sandbox>

For each ICS we want to run
    For each student submission
        safely run it
    Potentially combine results

Config settings
    seed
    create_missing_outputs
"""

import argparse
from pedal.command_line.modes import MODES
from pedal.environments import ALL_ENVIRONMENTS


def main(args=None):
    """
    Actually runs Pedal from the command line.

    Args:
        args (argparse.Namespace): The arguments parsed from the command line.
    """
    # Get command line arguments, unless we were explicitly given them.
    if args is None:
        args = parse_args()
    pipeline = MODES.PIPELINES[args.mode]
    return pipeline(args).execute()


def parse_args(reduced_mode=False):
    """ Parse the arguments passed into the command line. """
    parser = argparse.ArgumentParser(description='Run instructor control '
                                                 'script on student submissions.')
    parser.add_argument('mode', help="What kind of Pedal analysis you're running",
                        choices=list(MODES.PIPELINES))
    if not reduced_mode:
        parser.add_argument('instructor', help='The path to the instructor control '
                                               'script, or multiple scripts.')
    parser.add_argument('submissions', help='The path to the student submissions.'
                                            ' Defaults to a folder named '
                                            'submissions adjacent to the '
                                            'instructor control script.',
                        default='submissions',
                        nargs='?')
    parser.add_argument("--alternate_filenames",
                        help="A semicolon separated list of potential filenames to try if the main isn't found.",
                        default="")
    # TODO: Handle output to file
    parser.add_argument('--output', '-o',
                        help='The output file path for the result. Defaults to stdout.',
                        default='stdout')
    parser.add_argument('--config', '-c',
                        help="Uses the configuration file to get settings.")
    parser.add_argument('--create_output', '-m',
                        help="In verify mode, creates any missing outputs.",
                        action='store_true')
    parser.add_argument('--environment', '-e',
                        help="Sets the environment context for this script, which"
                             " can run special setups and override tools as"
                             " needed.", default='standard', choices=ALL_ENVIRONMENTS)
    parser.add_argument('--instructor_name',
                        help="Sets the name of the instructor file to something"
                             " more friendly. If not given, then will default"
                             " to the instructor filename.", default=None)
    parser.add_argument('--progsnap_profile',
                        default='blockpy', # TODO: Change this biased default
                        help="Uses the given profile's default settings for"
                             " loading in a ProgSnap2 dataset",
                        )
    parser.add_argument('--include_scripts', help='An optional filter to '
                                                  'only include certain scripts',
                        default=None)
    parser.add_argument('--limit', help='An optional limit to how many'
                                        ' submissions are run. Mostly for testing'
                                        ' purposes.',
                        default=None)
    parser.add_argument('--resolver', help='Choose a different resolver to use (the name of a function defined in the instructor control script).',
                        default='resolve')
    parser.add_argument('--ics_direct', help="Give the instructor code directly"
                                             " instead of loading from a file.",
                        default=False, action='store_true')
    # TODO: Elegant way of skipping specific, common phases
    parser.add_argument('--skip_tifa', help="Skip using TIFA in the environment",
                        default=False, action='store_true')
    parser.add_argument('--skip_run', help="Skip automatically running student code in the environment",
                        default=False, action='store_true')
    parser.add_argument('--progsnap_events', help="Choose what level of event"
                                                  " to capture from Progsnap event"
                                                  " logs.",
                        default='run', choices=['run', 'edit', 'last', 'compile'])
    parser.add_argument('--cache', help='Use the given directory to hold the cache.'
                                        ' You can use "./" to use the current directory.',
                        default=False)
    # TODO: Want to allow for multiple threads too to parallel process data (faster!)
    parser.add_argument('--threaded', help='Run the instructor script in a separate thread to avoid'
                                           ' timeout crashes.',
                        default=False)
    parser.add_argument('--log_level', help="Set the logging level for Pedal.",
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                        default="ERROR")
    '''
    parser.add_argument('--include_submissions', help='An optional REGEX filter '
                                                      'to only include certain submissions')
    parser.add_argument('--exclude_submissions', help='An optional REGEX filter '
                                                      'to remove certain submissions')
    parser.add_argument('--include_scripts', help='An optional REGEX filter to '
                                                  'only include certain scripts')
    parser.add_argument('--exclude_scripts', help='An optional REGEX filter to '
                                                  'remove certain scripts')
    parser.add_argument('--parallel_scripts', help="Which style to use for "
                                                   "running scripts in parallel.",
                        choices=["threads", "processes", "none"])
    '''
    args = parser.parse_args()
    if args.instructor_name is None:
        args.instructor_name = args.instructor
    return args


if __name__ == '__main__':
    main()

# If scripts is a single python file
# If scripts is a folder
# If scripts is a zip file
# If scripts is a ProgSnap SQL file



"""
Tests related to checking function definitions
"""

import unittest
from dataclasses import dataclass

from pedal import unit_test, block_function, set_source, clear_report
from pedal.assertions.static import *
from pedal.assertions.commands import *
from pedal.assertions.runtime import *
from pedal.resolvers import simple
from pedal.types.new_types import DictType, LiteralStr, StrType
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE


def correct(a, b):
    if not b:
        return "0%"
    return str(100 * a // b) + "%"


def no_b_check(a, b):
    return str(100 * a // b) + "%"


def really_bad(a, b):
    return "0%"


class TestAssertionsWheatChaffGame(ExecutionTestCase):
    def test_wheatcaff_game_valid(self):
        clear_report()
        set_source("""
from bakery import assert_equal

assert_equal(divide(1, 1), "100%")
""")
        game = wheat_chaff_game('divide', {'correct': correct}, {})
        self.assertTrue(game)

    def test_wheatcaff_game_not_valid(self):
        clear_report()
        set_source("""
from bakery import assert_equal

assert_equal(divide(1, 1), "30%")
        """)
        game = wheat_chaff_game('divide', {'correct': correct}, {})
        self.assertFalse(game)

    def test_wheatcaff_game_thorough(self):
        clear_report()
        set_source("""
from bakery import assert_equal

assert_equal(divide(1, 0), "0%")
""")
        game = wheat_chaff_game('divide', {}, {'bad': no_b_check})
        self.assertTrue(game)

    def test_wheatcaff_game_not_thorough(self):
        clear_report()
        set_source("""
from bakery import assert_equal

assert_equal(divide(1, 1), "100%")
        """)
        game = wheat_chaff_game('divide', {}, {'bad': no_b_check})
        self.assertFalse(game)

    def test_wheatchaff_game_thorough_valid(self):
        clear_report()
        set_source("""
from bakery import assert_equal

assert_equal(divide(1, 1), "100%")
assert_equal(divide(1, 0), "0%")
        """)
        game = wheat_chaff_game('divide', {'correct': correct}, {'bad': no_b_check})
        self.assertTrue(game)

    def test_wheatchaff_game_not_thorough_not_valid(self):
        clear_report()
        set_source("""
from bakery import assert_equal

assert_equal(divide(1, 1), "0%")
        """)
        game = wheat_chaff_game('divide', {'correct': correct}, {'bad': really_bad})
        self.assertFalse(game)
        final = simple.resolve()
        self.assertEqual("""
I ran your test cases against some of my own implementations of divide.
I had 1 programs I expected to pass, and 1 programs I expected to fail.
Your tests did not pass 1 of my good programs.
Your tests did not catch 1 of my bad programs.

Here are the names for the good programs that incorrectly failed on your test cases:
    correct
Here are the names for the bad programs that incorrectly passed your tests:
    bad""", final.message.rstrip())

    def test_wheatchaff_game_types_wrong(self):
        clear_report()
        set_source("""
from bakery import assert_equal

assert_equal(divide(1, "1"), "0%")
        """)
        game = wheat_chaff_game('divide', {'correct': correct}, {'bad': really_bad},
                                type_signature=(int, int, str))
        self.assertTrue(game)
        final = simple.resolve()
        self.assertEqual("""I ran your test cases against some of my own implementations of divide.
At least one of your tests had the wrong parameter or return type.
Make sure that all of your tests have valid types.""", final.message.rstrip())

if __name__ == '__main__':
    unittest.main(buffer=False)



import json
from pedal.core.feedback import Feedback
from pedal.core.commands import set_correct
from pedal.core.report import Report
from pedal.core.scoring import Score, combine_scores


def parse_feedback(feedback):
    """

    Args:
        feedback:

    Returns:

    """
    message = feedback.message
    title = feedback.title or feedback.label
    return feedback.correct, feedback.score, message, title, feedback.fields


class FinalFeedback:
    """
    Internal class used for organizing the feedback into one place.
    Dumpable into a simple dictionary.
    """

    DEFAULT_NO_FEEDBACK_TITLE = "No Errors"
    DEFAULT_NO_FEEDBACK_MESSAGE = "No errors reported."
    DEFAULT_NO_FEEDBACK_LABEL = "set_correct_no_errors"

    # TODO: Change all these so that FinalFeedback also logs considered feedback

    def __init__(self, correct=None, score=None, category=None, label=None, title=None,
                 message=None, data=None, hide_correctness=None,
                 suppressions=None, suppressed_labels=None, success=None):
        """

        Args:
            correct:
            score:
            category:
            label:
            title:
            message:
            data:
            hide_correctness:
            suppressions:
            suppressed_labels:
            success (bool): DEPRECATED. This field is now set by `correct`. The alternative parameter name is
                kept around for backwards compatibility.
        """
        self.success = self.correct = correct if correct is not None else success
        self.score = score
        self._scores = []
        self._scores_feedback = []
        self.category = category
        self.label = label
        self.title = title
        self.message = message
        self.data = data
        self.hide_correctness = hide_correctness
        self.positives = []
        self.instructions = []
        self.systems = []
        self.suppressions = suppressions
        self.suppressed_labels = suppressed_labels
        self.considered = []
        self.used = []

    def merge(self, feedback):
        self.considered.append(feedback)
        category = feedback.category.lower()
        if category in self.suppressions:
            if True in self.suppressions[category]:
                return
            else:
                looking_for = feedback.label.lower()
                if looking_for in self.suppressions[category]:
                    list_of_fields = self.suppressions[category][looking_for]
                    # Go through each of the sets of fields
                    for fields in list_of_fields:
                        # Do any of them NOT match?
                        for field, value in fields.items():
                            if feedback.fields.get(field, None) != value:
                                break
                        else:
                            # Oh hey, a match, let's skip this guy
                            return
        if feedback.label in self.suppressed_labels:
            list_of_fields = self.suppressed_labels[feedback.label]
            # Go through each of the sets of fields
            for fields in list_of_fields:
                # Do any of them not match?
                for field, value in fields.items():
                    if feedback['fields'].get(field, None) != value:
                        break
                else:
                    return
        correct, partial, message, title, data = parse_feedback(feedback)
        if feedback and feedback.category == Feedback.CATEGORIES.SYSTEM:
            self.systems.append(feedback)
        if not feedback.unscored and feedback.score is not None:
            # Triggered Negative leads to opposite behavior for operator
            # Also untriggered positive feedback
            invert_logic = ((feedback.valence != feedback.NEGATIVE_VALENCE) == (not feedback))
            inversion = "!" if invert_logic else ""
            self._scores.append(f"{inversion}{partial}")
            feedback.resolved_score = Score.parse(f"{inversion}{partial}").to_percent_string()
        if not feedback and feedback.else_message:
            self.positives.append(feedback)
            return feedback
        if not feedback or feedback.muted:
            return
        if feedback.kind == Feedback.KINDS.COMPLIMENT:
            self.positives.append(feedback)
            return feedback
        if feedback.kind == Feedback.KINDS.INSTRUCTIONAL:
            self.instructions.append(feedback)
        self.success = self.correct = correct and self.correct
        if message is not None and self.message is None:
            self.message = message
            self.title = title
            self.category = feedback.category
            self.label = feedback.label
            self.data = data
            self.used.append(feedback)
        return feedback

    def finalize(self):
        if self.message is None:
            self.title = self.DEFAULT_NO_FEEDBACK_TITLE
            self.message = self.DEFAULT_NO_FEEDBACK_MESSAGE
        self.hide_correctness = self.suppressions.get('correct', self.suppressions.get('success', False))
        if (not self.hide_correctness and
                self.label == self.DEFAULT_NO_FEEDBACK_LABEL and
                self.category == Feedback.CATEGORIES.COMPLETE):
            # TODO: Promote to be its own atomic feedback function
            self.title = set_correct.title
            self.message = set_correct.message_template
            self.score = 1
            self.success = self.correct = True
        else:
            self.score = combine_scores(self._scores)
        self.success = self.correct = bool(self.correct)
        return self

    def __str__(self) -> str:
        return "FinalFeedback({label!r}, {title!r}, {message!r})".format(label=self.label,
                                                                         title=self.title,
                                                                         message=self.message[:50])

    def for_console(self) -> str:
        return "{label}\n{score}\n{title}\n{message}".format(label=self.label,
                                                             score=self.score,
                                                             title=self.title,
                                                             message=self.message)

    def to_json(self) -> dict:
        """

        Returns:

        """
        return {
            'correct': self.correct,
            'success': self.correct,
            'score': self.score,
            "scores": self._scores,
            'category': self.category,
            'label': self.label,
            'title': self.title,
            'message': self.message,
            'data': self.data,
            'hide_correctness': self.hide_correctness,
            'location': self.data.get('location', None) if self.data else None
        }

    def to_file(self, path, format='json'):
        with open(path, 'w') as f:
            if format == 'json':
                json.dump(self.to_json(), f)
            else:
                f.write(self.for_console())

def set_correct_no_errors(report: Report) -> FinalFeedback:
    return FinalFeedback(correct=True, score=0,
                         title=None, message=None,
                         category=Feedback.CATEGORIES.COMPLETE,
                         label=FinalFeedback.DEFAULT_NO_FEEDBACK_LABEL,
                         data=[], hide_correctness=False,
                         suppressions=report.suppressions,
                         suppressed_labels=report.suppressed_labels)



"""
Unifying collection of all the commands in ``pedal.assertions``.
"""
from pedal.core.commands import gently
from pedal.core.scoring import Score, combine_scores
from pedal.sandbox.commands import call, clear_student_data, run, get_sandbox, CommandBlock
from pedal.assertions.static import *
from pedal.assertions.runtime import *
from pedal.assertions.runtime import _FIELDS
from pedal.assertions.positive import close_output, correct_output
from pedal.types.new_types import is_subtype
from pedal.assertions.testing_libraries import *

try:
    from dataclasses import fields
except:
    fields = None


class unit_test(assert_group):
    justification_template = ("Some of the feedback in this group was triggered.",
                              "None of the feedback in this group was triggered.")

    def __init__(self, function_name, given_partial_credit=None, given_score=None, **kwargs):
        super().__init__(function_name, **kwargs)
        self.given_partial_credit = given_partial_credit
        self.given_score = given_score

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        if self.given_partial_credit is False:
            self.score = self.given_score
        else:
            self.valence = self.POSITIVE_VALENCE
            self.score = combine_scores([success.score for success in self.successes],
                                        [failure.score for failure in self.failures] +
                                        [error.score for error in self.errors])
        return False


_unit_test_class = unit_test


def unit_test(function, *tests, else_message=None, else_message_template=None,
              score=None, partial_credit=False,
              assert_function=None, context=None, **kwargs):
    """
    Helper function for quickly unit testing.

    * Specify exact score for each test case, individually | list[str]
    * Specify exact score for each test case, with single value | str
    * Specify total score for all test cases, no partial credit | False, using `score`
    * Specify total score for all test cases, divide by # of cases passed | True, using `score`

    Args:
        assert_function: Override the assert function (defaults to assert_equal)
        function (str): The name of the function to test.
        *tests (): The test cases to run. Each test case is a tuple of the input/output pairs.
        else_message (str): The message to present as a positive if the
             tests all pass.
        else_message_template (str): The template for the else_message, to be formatted.
        score (str): The score to give overall/per test case, depending on
            partial_credit.
        partial_credit (bool or str or list[str]): Whether to give credit for
            each unit test as a percentage of the whole (True), or to only give
            points for passing all the tests (False). Defaults to False.
            If a list is passed, those scores will be used per test case.
        context (Sandbox or CommandBlock): The context to run the function in.
        **kwargs ():

    Returns:

    """
    if assert_function is None:
        assert_function = assert_equal
    each_score = partial_credit_logic(tests, score, partial_credit)
    # Handle context
    if context is True:
        context = get_sandbox().get_context()
    elif context:
        if isinstance(context, Sandbox):
            context = context.get_context()
        elif isinstance(context, CommandBlock):
            context = context.context
        else:
            context = get_sandbox().get_context(context._actual_context_id)
    unit_test_context = _unit_test_class(function,
                                         else_message=else_message,
                                         else_message_template=else_message_template,
                                         context=context, given_partial_credit=partial_credit,
                                         given_score=score, **kwargs)
    if not tests:
        return unit_test_context
    # TODO: Make it so unit_test can document its cases
    with unit_test_context as group_result:
        for test_index, test in enumerate(tests):
            args, expected = test
            if isinstance(args, str):
                assert_function(call(function, args_locals=[args]), expected, score=each_score[test_index], **kwargs)
            else:
                assert_function(call(function, *args), expected, score=each_score[test_index], **kwargs)

    return not group_result


class check_dataclass_error(RuntimeAssertionFeedback):
    justification = "Value is not a dataclass"
    _expected_verb = "to be"
    _inverse_operator = "is not"

    def __init__(self, value, **kwargs):
        super().__init__(SandboxedValue(value), ExactValue("a dataclass"), **kwargs)

def check_dataclass_instance(value, dataclass, else_message=None, score=None, partial_credit=False, **kwargs):
    report = kwargs.get('report', MAIN_REPORT)
    name = dataclass.__name__
    # Check that we weren't given an error
    if isinstance(value, Exception):
        return check_dataclass_error(value, score=score, **kwargs)
        #return gently(f"I evaluated the name {report.format.name(name)} and expected the result to be a dataclass. "
        #              f"However, there was an error instead.",
        #              label="check_dataclass_error", score=score, **kwargs)
    # Get the students' instance as a Pedal Type
    value_pedal_type = get_pedal_type_from_value(unwrap_value(value), evaluate)
    # Convert the instructor's version to a Pedal type
    evaluated_expected_type = evaluate(dataclass) if isinstance(dataclass, str) else dataclass
    expected_pedal_type = normalize_type(evaluated_expected_type, evaluate)
    if not isinstance(expected_pedal_type, Exception):
        expected_pedal_type = expected_pedal_type.as_type()
        expected_pedal_type_name = expected_pedal_type.singular_name
    else:
        expected_pedal_type_name = ""
    # Make sure the students' version is an actual dataclass
    singular_name = share_sandbox_context(value_pedal_type.singular_name, value)
    if not hasattr(unwrap_value(value), _FIELDS):
        return check_dataclass_error(value, score=score, **kwargs)
        #return gently(f"I evaluated the name {report.format.name(name)} and expected the result to be a dataclass. "
        #              f"However, the result was a {report.format.python_expression(str(type(unwrap_value(value))))}",
        #              label="check_dataclass_missing", score=score, **kwargs)
    if not fields:
        return gently("Dataclasses are not supported in this version of Python", label="dataclasses_not_supported")
    # Now Check the fields
    actual_fields = {field.name: get_pedal_type_from_value(getattr(unwrap_value(value), field.name), evaluate) for field
                     in fields(unwrap_value(value))}
    expected_fields = {field.name: field.type for field in fields(dataclass)}
    # Number of fields
    expected_arity = len(expected_fields)
    actual_arity = len(actual_fields)
    if actual_arity < expected_arity:
        return too_few_fields(name, actual_arity, expected_arity, **kwargs)
    elif actual_arity > expected_arity:
        return too_many_fields(name, actual_arity, expected_arity, **kwargs)
    # Checks each field's name and type
    for actual_field_name, actual_field_type in actual_fields.items():
        if actual_field_name not in expected_fields:
            return unknown_field(name, actual_field_name, **kwargs)
        expected_field_type = expected_fields[actual_field_name]
        expected_field_type = normalize_type(expected_field_type, evaluate).as_type()
        try:
            actual_field_type = normalize_type(actual_field_type, evaluate).as_type()
        except ValueError as e:
            return invalid_field_type(name, actual_field_name, actual_field_type, expected_field_type, **kwargs)
        if not is_subtype(actual_field_type, expected_field_type):
            return wrong_fields_type(name, actual_field_name, actual_field_type,
                                     expected_field_type, **kwargs)
    return None


class wheat_chaff_game_class(assert_group):
    message_template = """
I ran your test cases against some of my own implementations of {name:name}.
I had {wheat_count} programs I expected to pass, and {chaff_count} programs I expected to fail.
{wheat_outcome}{chaff_outcome}
{wheat_names}{chaff_names}
"""

    def __init__(self, name, wheats, chaffs, **kwargs):
        super().__init__(name=name, **kwargs)
        self.wheats = wheats
        self.correct_wheats = []
        self.chaffs = chaffs
        self.correct_chaffs = []
        self.points = 0
        self.points_possible = 2
        self.fields['name'] = name
        self.fields['wheat_count'] = len(wheats)
        self.fields['chaff_count'] = len(chaffs)

    def __exit__(self, exc_type, exc_value, traceback):
        self.report.stop_group(self)
        self.format_message()
        self._handle_condition()
        return False

    def format_message(self):
        missing_wheats = len(self.wheats) - len(self.correct_wheats)
        if missing_wheats:
            self.fields['wheat_outcome'] = f"Your tests did not pass {missing_wheats} of my good programs.\n"
            names = "\n".join([name for name in self.wheats.keys() if name not in self.correct_wheats])
            self.fields[
                'wheat_names'] = "Here are the names for the good programs that incorrectly failed on your test cases:\n" + self.report.format.output(
                names) + "\n"
        else:
            self.fields['wheat_outcome'] = "Your tests passed all of my good programs.\n"
            self.fields['wheat_names'] = ""
        missing_chaffs = len(self.chaffs) - len(self.correct_chaffs)
        if missing_chaffs:
            self.fields['chaff_outcome'] = f"Your tests did not catch {missing_chaffs} of my bad programs.\n"
            names = "\n".join([name for name in self.chaffs.keys() if name not in self.correct_chaffs])
            self.fields[
                'chaff_names'] = "Here are the names for the bad programs that incorrectly passed your tests:\n" + self.report.format.output(
                names)
        else:
            self.fields['chaff_outcome'] = "At least one of your tests correctly failed for all of my bad programs.\n"
            self.fields['chaff_names'] = ""

    def wheat_outcome(self, wheat, success):
        if success:
            self.points += 1 / len(self.wheats)
            self.correct_wheats.append(wheat)

    def chaff_outcome(self, chaff, failure):
        if failure:
            self.points += 1 / len(self.chaffs)
            self.correct_chaffs.append(chaff)



    def condition(self):
        correct = len(self.correct_wheats) + len(self.correct_chaffs)
        possible = len(self.wheats) + len(self.chaffs)
        return correct < possible


def partial_credit_logic(cases, score, partial_credit):
    if isinstance(partial_credit, bool):
        if partial_credit:
            # Specify total score for all test cases, divide by # of cases passed | True, using `score`
            if score:
                return [str(Score.parse(score) / len(cases)) for case in cases]
            else:
                return [None for case in cases]
        else:
            # Specify total score for all test cases, no partial credit | False, using `score`
            return [None for case in cases]
    elif isinstance(partial_credit, (int, str, float)):
        # Specify exact score for each test case, with single value | str
        return [partial_credit for case in cases]
    else:
        # Specify exact score for each test case, individually | list[str]
        return partial_credit


def get_success_passed_failed_total_reason(student):
    student = get_student_data()
    if 'assert_equal' not in student:
        return False, 0, 0, 0, "Failed to import `assert_equal`"
    assert_equal = student['assert_equal']
    if not hasattr(assert_equal, 'student_tests'):
        return False, 0, 0, 0, "The `assert_equal` function has been modified."
    student_tests = assert_equal.student_tests
    return True, student_tests.successes, student_tests.failures, student_tests.tests, ''


def check_type_signature(function_name, function, type_signature, type_failures):
    if type_signature is None:
        return function

    def check_types(*args, **kwargs):
        if len(type_signature)-1 != len(args):
            type_failures.append(TypeError(f"Expected {len(type_signature)-1} arguments, but got {len(args)}"))
        for i, (arg, expected_type) in enumerate(zip(args, type_signature[:-1])):
            expected_type, _ = type_to_pedal_type(expected_type)
            arg = value_to_pedal_type(arg)
            if not is_subtype(arg, expected_type):
                type_failures.append(f"Argument {i} should be {expected_type}, but got {type(arg)}")
        result = function(*args, **kwargs)
        expected_type, _ = type_to_pedal_type(type_signature[-1])
        result_type = value_to_pedal_type(result)
        if not is_subtype(result_type, expected_type):
            type_failures.append(f"Expected return type of {expected_type}, but got {result_type}")
        return result

    return check_types


def wheat_chaff_game(function_name, wheats, chaffs,
                     else_message=None, score=None, partial_credit=True, type_signature=None,
                     **kwargs):
    report = kwargs.get('report', MAIN_REPORT)
    each_score = partial_credit_logic([0] * (len(wheats) + len(chaffs)), score, partial_credit)
    type_failures = []
    with wheat_chaff_game_class(function_name, wheats, chaffs, **kwargs) as group_result:
        sandbox = get_sandbox(report)
        last = 0
        for kind, iterator in [("wheat", wheats.items()), ("chaff", chaffs.items())]:
            for i, (name, wheat) in enumerate(iterator, start=last):
                clear_student_data()
                student = get_student_data(report)
                student[function_name] = check_type_signature(function_name, wheat, type_signature, type_failures)
                result = run(report=report, before='from bakery import assert_equal\nassert_equal.student_tests.reset()')
                success, passed, failed, total, reason = get_success_passed_failed_total_reason(result)
                if kind == 'wheat':
                    group_result.wheat_outcome(name, not sandbox.exception and not failed and success and passed)
                else:
                    group_result.chaff_outcome(name, sandbox.exception or failed or not success)
                if 'assert_equal' in student:
                    student['assert_equal'].student_tests.reset()
                last = i

    if type_failures:
        return gently(f"""I ran your test cases against some of my own implementations of {function_name}.
At least one of your tests had the wrong parameter or return type.
Make sure that all of your tests have valid types.""", label='wheat_chaff_game_types', title='Invalid Type for Test', **kwargs)
    if partial_credit is False:
        group_result.score = score
    else:
        group_result.valence = group_result.POSITIVE_VALENCE
        group_result.score = group_result.points / group_result.points_possible
    return not group_result
